/*
 *   	 Robinson-Schensted Correspondence, Yong Diagram, Weyl group
 *
 *					by Toshio OSHIMA
 
    Letter: 1,2,3,...,9,a,b,c,d,...
	 where a=10, b=11, c=12,...

	An element  w  of the Weyl group  W  can be represented by  <w>:
		<w> = w<e> with <e> = (1,2,...,n-1,n)

	A_{n-1}:	W = <s_1,...,s_{n-1}>
	B_n, C_n:	W = <s_1,...,s_{n-1},s_n>
	D_n:		W = <s_1,...,s_{n-1},s_n'>

	Suppose <w> = (a_1,...,a_n):

	s_i with 0<i<n is the permutation (i,i+1):
    	s_i<w>  =  (a_1,...,a_{i+1},a_i,...,a_n)

	s_n changes the sign of a_n (modulo MOD):
		s_n<w>  =  (a_1,...,a_{n-1},MOD-a_n)

	s_n' changes the signs of a_{n-1} and a_n (modulo MOD) and permute them.
		s_n'<w> =  (a_1,...,MOD-a_n,MOD-a_{n-1})

    n = rank(+1) should be smaller than MOD0 = MOD/2   (= 64)
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<math.h>

// #define	NOW

#define		ERROR		-1
#define		MAX_rank	63
#define		MAX_LEN		4000		/* Maximal length of the element */
#define		MAX_BUFW	0x100000	/* 1 M byte for elements of Weyl group */
#define		MAX_POWER	30000
#define		MAX_SORT	10000
#define		MOD			128			/* Should be larger than 4*rank */
#define		MOD0		64			/* = MOD/2	*/
#define		MAX_SIZE	64

#define		is_pos(x)	((x)<MOD0)
#define		is_neg(x)	((x)>MOD0)
#define		is_eof(x)	(!(x))
#define		to_abs(x)	((x)<=MOD0?(x):(MOD-(x)))
#define		to_neg(x)	(MOD-(x))
#define		to_pack(x)	(is_neg(x)?((x)+rank2-MOD):(x))
#define		to_unpack(x)	(((x)>rank)?((x)+MOD-rank2):(x))
#define		type_bc		(wtype&1)

int			rank;				/* Rank (+1 for A_n) of Weyl group  */
int			rank2;				/* 2*rank - 1 */
int			siz0;				/* siz0 = rank-1 */
int			siz1;				/* rank-1 if An, rank if Bn, Cn or Dn */
int			wtype;				/* Type of Weyl group */
								/* 0: An, 1: Bn, 2: Dn, 3:Cn */
int			w_max;				/* maximal number for w_buf  */
int			w_count;			/* number of elements in w_buf */
int			count;				/* counter for working */
int			f_debug;

#define	F_TABL		1
#define	F_RANK		2
#define	F_SHORT		4
#define	F_TWO		8
#define	F_ALL		0x10
#define	F_ORD		0x20
#define	F_ROOT		0x40
#define	F_NEG		0x80
#define	F_MUL		0x100
#define	F_LINEAR	0x200
#define	F_GETTAB	0x400
#define	F_LONG		0x800
#define	F_CELL		0x1000
#define	F_2CELL		0x2000
#define	F_OPTION	0x4000
#define	F_NOPRINT	0x8000
#define	F_ORD0		0x10000
#define	F_ORD1		0x20000
#define	F_COSET		0x40000
#define	F_PDECOMP	0x80000

#define	ER_OK		0
#define	ER_LONG		-2
#define	ER_BADYD	-3
#define	ER_BADWY	-4
#define	ER_SYSTEM	-5
#define	ER_RANK		-6
#define	ER_BADYT	-7
#define	ER_WBUFOVER	-8
#define	ER_TBUFOVER	-9
#define	ER_BIGRANK	-10
#define	ER_DIFRANK	-11
#define	ER_CELLA	-12
#define	ER_EXCEPT	-13
#define	ER_BADRANK	-14

					/* Value of Debug Option (f_debug) */
#define	FD_FUNDWEIGHT	0x01
#define	FD_RHOSHIFT		0x02
#define	FD_SHIFTWEIGHT	0x04
#define	FD_HOOKLENGTH	0x08
#define	FD_RS			0x10
#define	FD_IRS			0x20
#define	FD_NEXTYT		0x40


char *msg_debug =
	"-debug[?X][value]\n\n"
	"value: X1a (in Hex) = 01011  means 2^1+2^3+2^4 (1,3,4 are ON)\n"
    "       -1 (default)\n"
	" 0: Fudamental Weight\n"
	" 1: Rho-Shift\n"
	" 2: Shift-Weight\n"
	" 3: Hook Length\n"
	" 4: Robinson-Schensted\n"
	" 5: Inverse of Robinson-Schensted\n"
	" 6: Next Young Tableau\n"
;

char *msg_err[] =
{
	"",
	"",
	"Too long expression",
	"Wrong expression of Youg Diagram",
	"Wrong expression of an element of Weyl group",
	"Internal error",
	"Wrong expression of rank of an element of Weyl group",
	"Wrong Young Tableau",
	"Too many elements of Weyl group appear",
	"Too many Young tableau appear",
	"Too big rank",
	"The elements have different rank",
	"Cell calculation is only supported for TYPE_A",
	"Support only Weyl group of classical type",
	"Rank is not correct"
};

#define	MAX_rRANK	15

#define	TYPE_A		0
#define	TYPE_B		1
#define	TYPE_C		3
#define	TYPE_D		2
#define	TYPE_E		4
#define	TYPE_F		5
#define	TYPE_G		6

typedef signed char*	WEYL;
typedef	signed char		AWEYL;
typedef long			PACK;
typedef signed char*  	YT;
typedef signed char*	YD;
typedef	signed char*	ROOTS;
typedef	signed char		UNIT;
typedef signed short	POWER;

int 	f_mode;
int 	max_power;			/* maximal number used in power[] */
int		top_len;			/* shortest length in w_buf_pt */
int		last_len;			/* longest length in w_buf_pt */

UNIT	r_exp[MAX_LEN+2];	/* Buffer for a reduced expression */
UNIT	w_buf[MAX_BUFW];	/* Buffer for working space */
int		w_buf_pt[MAX_LEN+2];	/* pointer of w_buf by length 	*/
POWER 	power[MAX_POWER];	/* Buffer for calculating dim */

char s_sep[2] = "/";

/*
 *  functions to calculate the product/quotient of integers
 *  To get/use the number, reduce() should be called after the last calculations.
 */
void fact(POWER *p, int n);			/* multiply <p> by <n>!	*/
void dfact(POWER *p, int n);		/* devide <p> by <n>! */
void reduce(POWER *p, int mx);		/* prime factorization <p> */
void clear_int(POWER *p, int mx);	/* clear the number <p> */
int check_int(POWER *p, int mx);	/* check if the number is an integer <p> */
long prod_power(POWER *p, int mx);	/* get the number as a long integer <p> */
long fprod_power(POWER *p, int mx, float *f);	/* get the number as a long integer <p> */
int prtprod_power(POWER *p, int mx, int keta, char *buf);
void print_power(POWER *p, int mx);	/* print the prime factorization of the number <p> */

int check_wy(WEYL);
int check_yd(YD);
int check_yt(YD, YT);

/*
 *  Get the longest element <w> in the Weyl group
 */
void longest_w(WEYL w);

/*
 *  Get an element <w> related to a parabolic subgroup of the Weyl group
 *  <p> represents the partition corresponding to the parabolic subgroup.
 *    <mode>  0: the longest element of the parabloc subgroup.
 *            1: the above element * the longest element of the Wely group
 *            2: the inverse of the above element
 *			  3: product by simple reflections
 *            4: inverse of the element
 */
void parab_w(WEYL w, UNIT *p, int mode);

/*
 *  Get the inverse of the given element of the Weyl group
 *  <u> = <w>^{-1} 
 */
void inv_w(WEYL u, WEYL w);

/*
 *  Calculate the product of the elements of the Weyl group
 *  <u> = <v> * <w>
 */
void mul_w(WEYL u, WEYL v, WEYL w);

/*
 *  Calculate the length of the Weyl group <w> and return it.
 */
int	length(WEYL w);

/*
 *  Compare two elements of the Weyl group with respect to the Bruhat ordering
 *  return  0:  <v> == <w>
 *          1:  <v> >  <w>
 *         -1:  <v> <  <w>
 *          2:  <v> <> <w>
 */
int comp_b(WEYL v, WEYL w);

/*
 *  Check if an element can be transformed to another by the reflection 
 *  of a root
 *
 *  In:  mode == 0  Bruhat order
 *            != 0  simple reflection by right
 *  ret  1:  Yes
 *       0:  No
 */
int isadj_rf(WEYL w, WEYL v, int mode);

/*
 *  List all the roots corresponding to the reflections which decrease the 
 *  length of the given element <w> of the Weyl group by just one.
 *  This function is used to list all the elemets of the Weyl group adjacent 
 *  to the given element <w> with respect to the Bruhat ordering.
 *  Each root represented by 2 byte in <root>.
 *  return:  number of roots x 2 (= used byte)
 */
int adj_br(WEYL w, UNIT *root);

/*  List all the simple roots whose reflections by right transform WEYL <w> 
 *  longer/shorter
 *
 *  In: mode  == 0 : longer
 *            != 0 : shorter
 *  ret: 2 X number of elements byte
 *       result is written in UNIT *p (NULL terminated)    
 */
int adj_srf(WEYL w, UNIT *p, int mode);

/*
 * Get reduced expressions of the element <w> of the Weyl group.
 *   mode    0: put all the reduced expression to stdout
 *           1: put a reduced expression to stdout
 *           2: get a reduced expression in <r_exp>.
 *   return:
 */
int get_r_exp(WEYL w, int mode);

/*
 *  List all the positive roots which change into negative by the action 
 *  of the given
 *  element <w> of the Weyl group.
 *  Each root represented by 2 byte in <root>.
 *    return:  number of roots x 2 (= used byte
 */
int chg_root(WEYL w, UNIT *root);

int unpack(WEYL, PACK, int);

/*
 *  Print an element <w> of the Weyl group or a tableau (partition) <w> as 
 *  a string
 */
void print_e(WEYL w);
void print_el(WEYL w, int len);

/*
 *  Print an element <w> of the Weyl group or a tableau (partition) <w> as a 
 *  series of integers of length <len>
 */
void print_ew(WEYL, int len);

void print_wn(WEYL w, int f_short);

/*
 * Print series of roots <root> 
 */
void print_r(UNIT *root);

/*
 * Print Young tableau (<tab>, <diag>) as a string
 */
void print_yt(YD diag, YT tab);

/*
 *  Calculate the Robinson-Schensted correspondence (<left>, <right>, <tab>) 
 *  to the given element <w> of the symmetric group.
 */
int rs(WEYL w, YT left, YT right, YD tab);

/*
 *  Give the element <w> of the symmetric group associated to the pair of 
 *  young Tableau  (<left>, <right>, <tab>) under the Robinson-Schensted 
 *  correspondence.
 */
int rs2w(WEYL w, YT left, YT right, YD tab);

void add_rho(WEYL);

/*
 *  Calculate the dimension of the representation of the simple Lie group by 
 *  giving the coefficients <weight> of the fundamental weights.
 *  Prime factorization of the dimension is given by POWER <power>.
 *  return : the dimension (long).
 *           negative value -> error 
 */
long dim_rep2(UNIT *weight);
long dim_rep(UNIT *weight);

/* 
 *  Calculate the dimension of the representation of the symmetric group 
 *  expressed by Yong diagram <diag>.
 *  Prime factorization of the dimension is given by POWER <power>.
 *  return : the dimension (long).
 *           negative value -> error
 */
long dim_yd(YD diag);

/*
 *   Change the Young diagram (partition) <diag> to next one.
 *   The order is under the inverse lexico graphic order.
 *     6=5+1=4+2=4+1+1=3+3=3+2+1=3+1+1+1=2+2+2=2+2+1+1=2+1+1+1=1+1+1+1+1+1
 *   return  > 0 : the number of raws
 *             0 : the given <diag> is the last one.
 */
int next_yd(YD diag);

/*
 *  Get all the young tableau <tab>  associated to the given Young 
 *  diagram <diag>.
 *  The tableau is obtained by the lexicographic order.
 *  mode ==  0:  Get the first <tab> associated to <diag>
 *           1:  Get the next <tab>.  
 *               <tab> and <diag> should be the former <tab> and <diag> given by 
 *               this routine.  Then it is rewrited to the next <tab>. 
 *  return   0:  get the first/next <tab>.
 *          -1:  the last <tab> is already given
 */
int next_yt(YT tab, YD diag, int mode);

/*
 *  Get next series <s> of non-negative integers of length <len> whose sum 
 *  is fixed
 *  It comes from the inverse of the lexicograpic order
 *     300 210 201 120 111 102 030 021 012 003
 *  return   0:  <s> is changed into the next one
 *          -1:  the last one is given */
int next_sum(UNIT *s, int len);

/*
 *  Permute w[] into the adjacent larger element under the lexicographic order
 *  w[0],...,w[rank-1]: given numbers (some numbers may be same)
 *
 *  return  0:  permuted
 *         -1:  given one is the largest
 */
int NextOrd(WEYL w);

/*
 *  Generate representatives (with minimal length) of the right cosets by a 
 *  parabolic subgroup
 *  YD <y>   partition (parabolic)
 *           NULL: next representative
 *  return   1: get the next representative
 *           0: end
 */
int NextRep(WEYL w, YD y);

/*
 *   Add WEYL  <w>  to w_buf[]
 * 
 *   In:  WEYL <w>
 *   Out: ret   >=  the same <w> exists at <ret>-th position
 *              -1  add <w>
 *   w_count: the number of elements in w_buf[]
 */
int add_w_buf(WEYL w);

/*
 *  add_w_buf() + count the multiplicity for <w> in power[]
 *  
 */
int add_w_buf_count(WEYL w);

/*
 *         Sort elements in w_buf by length
 *
 *  The number of elements is given by w_count.
 *
 *  top_len, last_len, w_buf_pt[last_len - top_len + 1]  are set
 *  The elements with length <len>  (top_len <= <len> <= last_len)
 *  start at w_buf_pt[i-top_len] and end at w_buf_pt[i-top_len+1] - 1.
 */
void sort_by_length(void);

/*
 *  Calculate Bruhat poset for the sorted elements in w_buf[]
 *  result: i-th element > power[power[i]]-th,...,power[power[i+1]-1]-th
 *            (all the adjacent relation: i = 0,...,w_count-1)
 *  ret   0: OK
 *       -1: buffer overflow
 */
int bruhat_poset(void);

/*
 *  Add  Young tableau <rs> to w_buf[]
 *  ret:  >= 0  already exists
 *          -1  add it
 */
int add_yt(UNIT *rs);

void set_rank(int r);
void print_w(WEYL w);

/*
 *  Get the representative with the minimal length in the given right coset 
 *  by  a parabolic subgroup.
 *
 *  ret  1 : change into the element with the minimal length
 *       0 : the given element is of the minimal length
 */
int repright(WEYL w, YD y);

/*
 *  Get representatives of the double cosets by parabolic subgroups
 *  mode 0: the representatives with minimal length
 *       1: the representatives with maximal length 
 */
void dcoset(YT yl, YT yr, int mode);

/*
 *  Parabolic decomposition
 *  mode 0: by right multiplication of a parabolic subgroup
 *       1: by left  multiplication of a parabolic subgroup
 */
void ParaDecomp(WEYL w, YT y, int mode);

/* ------------------------------------------------------------------------ */

void err_exit(int id)
{
	if(id >= 0)
		return;
	id = -id;
	if(id > 0 && id < sizeof(msg_err)/sizeof(char *))
		fprintf(stderr, "%s!\n", msg_err[id]);
	exit(1);
}

/*
 *  Show Bruhat relation between sets of elements with adjacent length
 *
 *  In: WEYL <w0>  top of elements of original elements with the same length
 *      int  <n0>  number of the original elements
 *      WEYL <w1>  top of refered elements with the adjacent length
 *      int  <n1>  number of the refered elements
 */
void print_bo(WEYL w0, WEYL w1, int n0, int n1, int len)
{
	int i, j, k, f;
	WEYL wp, wq;

	for(wp = w0, i = 0; i < n0; wp += rank, i++){
		k = f = 0;
		for(wq = w1, j = 0; j < n1; wq += rank, j++){
			if(isadj_rf(wp, wq, f_mode & (F_ORD0|F_ORD1))){
				if(*s_sep){
					if(!f++)
						printf("%d:", i+1);
					printf((k++ > 0)?" %d":"%d", j+1);
				}else
					printf("(%d,%d)<(%d,%d)\n",i+1, len, j+1, len+1);
			}
		}
		if(f)
			printf("|");
	}
}

/*
 *   Add WEYL  <w>  to w_buf[]
 * 
 *   In:  WEYL <w>
 *   Out: ret   >=  the same <w> exists at <ret>-th position
 *              -1  add <w>
 */
int add_w_buf(WEYL w)
{
	int	i;
	UNIT *p;

	p = &w_buf[0];
	for(i = 0; i < w_count; p += rank, i++){
		if(!memcmp(p, w, rank))
			return i;
	}
	if(++w_count >= w_max)
		err_exit(ER_WBUFOVER);
	memmove(p, w, rank);
	return -1;
}

/*
 *  add_w_buf() + count the multiplicity for <w> in power[]
 */
int add_w_buf_count(WEYL w)
{
	int i;

	i = add_w_buf(w);
	if(i < 0){
		if(w_count < MAX_POWER-1)
		power[w_count-1] = 1;
	}else if(i < MAX_POWER-1)
		power[i]++;
	return i;
}

static int comp_lsub(const void *x, const void *y)
{
	return *(int *)x-*(int*)y;
}

/*
 *         Sort by length of elements in w_buf
 *
 *  The number of elements is given by w_count.
 *
 *  top_len, last_len, w_buf_pt[last_len - top_len + 1]  are set
 *  The elements with length <len>  (top_len <= <len> <= last_len)
 *  start at w_buf_pt[i-top_len] and end at w_buf_pt[i-top_len+1] - 1.
 *
 */
void sort_by_length(void)
{
	UNIT *w, *u, *tmp;
	int	i, j, len, *idx;

	if(w_count <= 1)
		return;

	idx = malloc(sizeof(int)*2*w_count);
	if(idx == NULL)
		return;
	for(w = &w_buf[0], i = 0; i < w_count; w += rank){
		idx[i*2] = length(w);						/* get index */
		idx[i*2+1] = i;
		i++;
	}
	qsort(idx, w_count, sizeof(int)*2, comp_lsub);	/* sort them */

	top_len = idx[0];								/* get bounds of length */
	last_len = idx[w_count*2-2];
	for(i = last_len - top_len + 2; i >= 0; i--)
		w_buf_pt[i] = 0;
	if(last_len - top_len >= MAX_LEN)
		goto quit;

	if((tmp = malloc(rank*w_count)) == NULL)		/* get tempolary buffer */
		goto quit;
	for(u = tmp, i = 0, len = -1; i < w_count; i++){	/* rearrange using it */
		w = w_buf + idx[2*i+1]*rank;
		for(j = 0; j < rank; j++)
			*u++ = w[j];
		if(len < idx[2*i]){
			len = idx[2*i];
			w_buf_pt[len-top_len] = i;
		}
	}
	memmove(w_buf, tmp, w_count*rank);
	free(tmp);

	j = last_len - top_len + 1;						/* fill length */
	w_buf_pt[j] = w_count;
	for(i = j, len = w_count; --i > 0; ){
		if(w_buf_pt[i] > 0 && w_buf_pt[i] < len)
			len = w_buf_pt[i];
		else
			w_buf_pt[i] = len;
	}
quit:
	free(idx);
}


/*
 *
 *
 */

void set_flg(char *flg, int k)
{
	int i, m, n;

	m = power[k];
	n = power[k+1];
	for(i = m; i < n; i++){
		if(!flg[power[i]]){
			flg[power[i]] = 1;
			set_flg(flg, power[i]);
		}
	}
}

/*
 *  Calculate Bruhat poset for the sorted elements in w_buf[]
 *  result: i-th element > power[power[i]],...,power[power[i+1]-1]
 *                (adjacent relation: i = 0,...,w_count-1)
 *  ret   0: OK
 *       -1: buffer overflow
 */
int bruhat_poset(void)
{
	int i, j, k, xx, yy, pt;
	char *flg;
	UNIT *w;

	if(w_count >= MAX_SORT)
		return -1;
	flg = (char *)malloc(w_count + 1);
	pt = w_count + 2;

	for(j= 0; j <= last_len - top_len; j++){
		xx = w_buf_pt[j];
		yy = w_buf_pt[j+1];
		for(i = xx; i < yy; i++){
			for(k = 0; k < i; k++)
				flg[k] = 0;
			power[i] = pt;
			w = w_buf + rank*i;
			for(k = xx-1; k >= 0; k--){
				if(comp_b(w, w_buf+ rank*k) == 1 && !flg[k]){
					power[pt++] = k;
					flg[k] = 1;
					set_flg(flg, k);
					if(pt >= MAX_POWER){
						free(flg);
						err_exit(ER_WBUFOVER);
						return -1;
					}
				}
			}
		}
	}
	power[i] = pt;
	free(flg);
	return 0;
}

/*
 *    Show elements of Weyl group recorded in w_buf[] after sorting them.
 *    If  (w_mode & F_ORD), then Bruhat relation is also shown
 */
void print_w_buf(int mode)
{
	UNIT *w;
	int i, j, k, xx, yy, m;

	sort_by_length();
	if(f_mode & F_NOPRINT)
		return;
	for(i = top_len, j= 0; i <= last_len; i++, j++){
			xx = w_buf_pt[j];
			yy = w_buf_pt[j+1];
			if( (k = yy-xx) <= 0 && (mode & 1) )
				continue;
			printf("\nlength %2d(%2d):", i, yy - xx);
			w = w_buf + xx*rank;
			for(m = 0; m < k; m++){
				print_el(w, rank);
				printf(" ");
				w += rank;
			}
			if(i < last_len && !(mode & 2) && (f_mode & F_ORD)){
				printf("\n");
				print_bo(w_buf + xx*rank, w_buf + yy*rank, yy-xx, w_buf_pt[j+2]-yy, i);
			}
	}
}

void print_wo(void)
{
	int i, j, k, len, num;
	UNIT *ww, *wp, w[MAX_rank], u[MAX_rank];

	num = 1;
	ww = w_buf + (w_max - num)*rank;
	for(i = 0; i < rank; i++)
		ww[i] = i+1;
	for(len = 1; ;len++){
		w_count = 0;
		for(i = 0, wp = ww; i < num; wp += rank, i++){
			for(j = 0; j < rank; j++)
				w[j] = wp[j];				/* copy the elemnt */
			for(k = 0; k < rank; k++){
				if(k < siz0){
					if(w[k] < w[k+1]){
						j = w[k];			/* transposition */
						w[k] = w[k+1];
						w[k+1] = j;
						add_w_buf(w);
						j = w[k];			/* recover the original */
						w[k] = w[k+1];
						w[k+1] = j;
					}
				}else{
					if(wtype == TYPE_B || wtype == TYPE_C){
						if(!is_neg(w[siz0])){			/* change sign */
							w[siz0] = to_neg(w[siz0]);
							add_w_buf(w);
							w[siz0] = to_neg(w[siz0]);
						}
					}else if(wtype == TYPE_D){			/* swap & change sign */
						if(w[siz0] + w[siz0-1] < MOD){
							j = w[siz0];
							w[siz0] = to_neg(w[siz0-1]);
							w[siz0-1] = to_neg(j);
							add_w_buf(w);
							j = w[siz0];
							w[siz0] = to_neg(w[siz0-1]);
							w[siz0-1] = to_neg(j);
						}
					}
				}
			}	/* end this <w> */
		}	/* end all <w> in this length */
		if(w_count + num >= w_max)
			err_exit(ER_WBUFOVER);
		if(f_mode & F_ORD)
			printf("\nlength %d(%d):", len, num);
		u[rank] = 0;
		for(wp = ww, i = 0; i < num; wp += rank, i++){
			if(f_mode & F_ORD){
				print_el(wp, rank);
				printf(" ");
			}else{
				for(j = 0; j < rank; j++)
					u[j] = wp[j];
				print_w(u);
				if(wtype == TYPE_A)
					printf("\n");
			}
		}
		if(f_mode & F_ORD)
			printf("\n");
		if(!w_count)
			break;
		if(f_mode & F_ORD)
			print_bo(ww, &w_buf[0], num, w_count, len);
		num = w_count;
		ww = w_buf + (w_max - num)*rank;
		memmove(ww, w_buf, num*rank);
	}
}

/* --------------------------------------------------------------------
 *           Fundamental functions for elements of Weyl group

   void mul_w(WEYL, WEYL, WEYL);
   void inv_w(WEYL, WEYL);
   PACK pack(WEYL);
   int unpack(WEYL, PACK, int);
   void longest_w(WEYL);
   void parab_w(WEYL, UNIT *, int mode);
   int length(UNIT *);
   int comp_b(WEYL, WEYL);
   int isadj_rf(WEYL, WEYL);
   int adj_br(UNIT *, UNIT *);
   int adj_srf(WEYL w, UNIT *p, int mode);
   int chg_root(WEYL, UNIT *)
   int get_r_exp(WEYL, int mode);
 * -------------------------------------------------------------------- */
/*
 *	Product of two elements: u = vw
 */
void mul_w( WEYL u, WEYL v, WEYL w )
{
	int		i, j;
	AWEYL	uu[MAX_SIZE];

	for( i = siz0; i >= 0; i-- ){
		if( (j = w[i]) < MOD0 )
			uu[i] = v[j-1];
		else uu[i] = MOD - v[(MOD-1)-j];
	}
	for( i = 0; i < rank ; i++ )
		u[i] = uu[i];
	u[i] = 0;
}

/*
 *	Inverse of an element: u = w^{-1}
 */
void inv_w( WEYL u, WEYL v )
{
	int		i, j;
	AWEYL	uu[MAX_SIZE];
	for( i = siz0; i >= 0; i-- ){
		if( (j = v[i]) < MOD0 ) uu[j-1] = i+1;
		else uu[(MOD-1)-j] = to_neg(i+1);
	}
	for(i = 0; i < rank; i++)
		u[i] = uu[i];
	u[i] = 0;
}

/*
 *   Calculate the position of the element <w> of a Weyl group 
 *   in a lexicographic order
 *
 *   In:   WEYL <w>
 *   Out:  the position
 */
PACK pack(WEYL w)
{
	int i, j, k, m, n;
	AWEYL s[MAX_SIZE], t[MAX_SIZE];

	if(wtype == TYPE_A){
		for(i = 0; i < rank; i++){
			n = w[i];
			for(k = 1, j = 0; j < i; j++){
				if(w[j] < n)
					k++;
			}
			s[i] = n - k;
		}
		k = rank-1;
		for(n = i = 0; i < k; i++)
			n += n*(rank-i) + s[i];
	}else{
		for(i = 0; i < rank; i++){
			m = w[i];
			t[i] = n = (is_neg(m))?to_neg(m):m;
			for(k = 1, j = 0; j < i; j++){
				if(t[j] < n)
					k++;
			}
			s[i] = is_neg(m)?(m+2*rank-MOD):(n-k);
		}
		k = (wtype == TYPE_D)?(rank-1):rank;
		for(n = i = 0; i < k; i++)
			n += 2*n*(rank-i) + s[i];
	}
	return n;
}

/*
 *     Get the <pk>-th element of a Weyl group 
 * 
 *  In:  <r>  the rank of the Weyl group
 *       <pk>
 *  Out:  ret  0:  WEYL <w>  the corresponding element
 *            -1:  error
 *
 *	non-negative integer <pk> -> element of Weyl group
 *  A_n:     			0 <=  <pk>  < rank!				 (n <= 11)
 *  B_n, C_n:           0 <=  <pk>  < 2^rank * rank!	 (n <=  9)
 *
 *  D_n:				0 <=  <pk>  < 2^{rank-1} * rank! (n <= 10)
 */
int unpack(WEYL w, PACK pk, int r)
{
	int i, j, k, l;
	AWEYL s[MAX_SIZE], t[MAX_SIZE];

	if(r > 255)
		return ERROR;
	s[1] = 0;
	if(wtype == TYPE_A){
		for(i = 2; i <= r; i++){
			s[i] = pk%i;
			pk = pk/i;
		}
		if(pk)
			return ERROR;
		for(i = 1; i <= r; i++)
			t[i] = i;
		w[r] = 0;
		for(i = r; i >= 1; i--){
			j = s[i];
			for(k = 1; j > 0; k++)
				if(t[k])
					j--;
			while(!t[k])
				k++;
			w[r-i] = t[k];
			t[k] = 0;
		}
	}else{
		i = (wtype == TYPE_D)?2:1;
		for(; i <= r; i++){
			s[i] = pk%(i*2);
			pk = pk/(i*2);
		}
		if(wtype == TYPE_D){
			for(k = 0, i = 2; i <= r; i++)
				if(s[i] >= i)
					k++;
				s[1] = (k & 1)?1:0;
		}
		if(pk)
			return ERROR;
		for(i = 1; i <= r; i++)
			t[i] = i;
		w[r] = 0;
		for(i = r; i >= 1; i--){
			j = s[i];
			l = (j < i)?j:(2*i - j - 1);
			for(k = 1; l > 0; k++)
				if(t[k])
					l--;
			while(!t[k])
				k++;
			w[r-i] = (j < i)?t[k]:to_neg(t[k]);
			t[k] = 0;
		}
	}
	return 0;
}

/*
 *    Get the longest element of the Weyl group
 */
void longest_w(WEYL w)
{
	int i;

	switch(wtype){
		case TYPE_A:
			for(i=0; i<rank; i++)
				w[i] = rank - i;
			break;
		case TYPE_B:
		case TYPE_C:
			for(i=0; i<rank; i++)
				w[i] = to_neg(i+1);
			break;
		case TYPE_D:
			for(i=0; i<rank; i++)
				w[i] = to_neg(i+1);
			if(rank & 1)
				w[i-1] = rank;
			break;
	}
	w[i] = 0;
}

/*
 *  Get an element <w> related to a parabolic subgroup of the Weyl group
 *  <p> represents the partition corresponding to the parabolic subgroup.
 *    <mode>  0: the longest element of the parabloc subgroup.
 *            1: the above element * the longest element of the Wely group
 *            2: the inverse of the above element
 *            3: product by simple reflection
 *            4: inverse of the element
 */
void parab_w(WEYL w, UNIT *p, int mode)
{
	int i, j, n, m;
	UNIT ww[MAX_rank], w2[MAX_rank];

	if(mode == 4){
		set_rank(strlen(p));
		inv_w(w,p);
		return;
	}
	if(mode == 3){
		if(!rank){
			for(i = rank = 0; p[i]; i++){
				if(p[i] > rank)
					rank = p[i];
			}
			if(wtype == TYPE_A)
				rank++;
			if(rank <= 0)
				rank = 1;
			set_rank(rank);
		}
		for(i = 0; i < rank; i++)
			ww[i] = i+1;
		for(i = 0; (j = p[i]) != 0; i++){
			if(j == rank)
				ww[siz0] = to_neg(ww[siz0]);
			else{
				m = ww[j-1];
				ww[j-1] = ww[j];
				ww[j] = m;
			}
		}
		ww[rank] = 0;
		strcpy(w, ww);
		return;
	}
	for(i = j = n = 0; (m = p[i++]) != 0 && m+n < MAX_rank; ){
		if(wtype == TYPE_A || p[i] != 0){
			for(j = n+m-1; j >= n; j--)		/* n+m, n+m-1,...,n+1 */
				ww[j] = 2*n+m-j;
			n += m;
		}else{
			m--;
			for(j = n+m-1; j >= n ; j--)
				ww[j] = to_neg(j+1);		/* -n-1, -n-2,...,-n-m+1 */
			n += m;
			if(wtype == TYPE_D && (m & 1))
				ww[n-1] = n;
		}
	}
	ww[n] = 0;
	set_rank(n);
	if(!mode){
		strcpy(w,ww);
	}else{
		longest_w(w2);
		if(mode == 1)
			mul_w(w,ww,w2);
		else
			mul_w(w,w2,ww);
	}
}

/*
 *	Get the length of an element of Weyl group
 */
int length( UNIT *w )
{
	int		i, j, len;

	for( i = len = 0; i < rank; i++ ){
		if( type_bc && is_neg(w[i]) ) len++;
		for( j = i+1; j < rank; j++ ){
			if( w[i] > w[j] ) len++;
			if( w[i] + w[j] > MOD ) len++;
		}
	}
	return( len );
}

/*
 *	Compare Bruhat ordering
 *	return value 0: equal, 1: w > v, -1: w < v, 2: no order between w and v
 */
int comp_b( WEYL w, WEYL v )
{
	int		wo[MAX_rank], vo[MAX_rank];
	int		i, j, w1, v1, flag, check;

	for( i = flag = check = 0; i < siz1; i++ ){
		check = 0;
		if( (w1 = w[i]) > (v1 = v[i]) ){
			if( flag < 0 ) 	check = 1;		/* should be checked */
			else if( !flag ) flag = 1;
		}
		else if( w1 < v1 ){
			if( flag > 0 )  check = 1;
			else if( !flag ) flag = -1;
		}
		for( j = i; --j >= 0 && wo[j] > w1; ) wo[j+1] = wo[j];
		wo[j+1] = w1;
		for( j = i; --j >= 0 && vo[j] > v1; ) vo[j+1] = vo[j];
		vo[j+1] = v1;
		if( check ){
			for( j = i; j >= 0; j-- ){
				if( ( w1 = wo[j] ) == ( v1 = vo[j] ) ) continue;
				if( w1 > v1 ){
					 if( flag < 0 ) return( 2 );
				}
				else if( flag > 0 ) return( 2 );
			}
		}
	}
	return( flag );
}

/*
 *  Check if an element can be transformed to another by the reflection 
 *  of a root
 *
 *  In:  mode == 0:      all reflections
 *            == F_ORD0: simple reflections by right
 *            == F_ORD1: simple reflection by left
 *  ret  1:  Yes
 *       0:  No
 */
int isadj_rf(WEYL u, WEYL v, int mode)
{
	int	i, k, l, count, uk;

	k = -1;
	for(i = count = 0; i < rank; i++){
		if(u[i] != v[i]){
			if(!count++)
				k = i;
			else{
				l = i;
				if(count > 2)
					return 0;
			}
		}
	}
	if(mode){
		if(mode & F_ORD1){
			uk = to_abs(u[k]);
			if(count <= 1){
				if(uk == rank)
					return 1;
				else return 0;
			}
			if(uk - to_abs(v[k]) != 1 && uk - to_abs(v[k]) != -1)
				return 0;
			if( (u[k] == v[l] && u[l] == v[k])
			  || (wtype == TYPE_D && (uk == siz0 || to_abs(u[l]) == siz0))
			      && u[k] == to_neg(v[l]) )
				return 1;
		}else{
			if(count <= 1){
				if(k == rank-1)
					return 1;
				else return 0;
			}
			if(k != l-1)
				return 0;
			if( (u[k] == v[l] && u[l] == v[k])
			  || (wtype == TYPE_D && l == siz0 && u[k] == to_neg(v[l])) )
				return 1;
		}
	}else{
		if(count <= 1
		  || (u[k] == v[l] && u[l] == v[k])
		  || (wtype == TYPE_D && u[k] == to_neg(v[l])) )
				return 1;
	}
	return 0;
}


/*
 *	Get elements which are smaller and adjacent w.r.t. Bruhat order
 *	Write the roots which represent the relations
 *	return value = 2*number of the elements
 *
 *	(k,l)w  < w  <->  a_k > a_l
 *  (k)w    < w  <->  a_k > MOD0
 *  (k,-l)w < w  <->  a_k + a_l > MOD
 *
 *
 *        Suppose       MOD0 = 6      MOD=12
 *
 *    The change the number of disordering pairs and pairs of negative sums
 *  1.      k    i    k                 k    i    k
 *       .......(9)..[2]..     -->   .......(3)..[2]..     (1-1) + (0-0) = 0
 *       ..[2]..(9).......     -->   ..[2]..(3).......     (0-0) + (0-0) = 0
 *       .......(9)..[a]..     -->   .......(3)..[a]..     (0-0) + (1-1) = 0
 *       ..[a]..(9).......     -->   ..[a]..(3).......     (1-1) + (1-1) = 0
 *       .......(9)..[4]..     -->   .......(3)..[4]..     (1-0) + (1-0) = 2
 *       ..[4]..(9).......     -->   ..[4]..(3)......      (0-1) + (1-0) = 0
 * 
 *    length((i)w) = length(w) - 1
 *        <=>  is_neg(a_i) && (i < k < rank => abs(a_i) > abs(a_k))
 *
 *  2.  s   i   t   j   u         s   i   t   j   u
 *    .[x].(4).[x].(a).[x]. --> .[x].(2).[x].(8).[x].
 *    .[x].(a).[x].(4).[x]. --> .[x].(8).[x].(2).[x].
 *     if x = 1, 5, 7, b, there exists no difference
 *
 *    .[3].(4).[3].(a).[3]. --> .[3].(2).[3].(8).[3].
 *          [s,i] + [s,j]        [(0-1)+(0-0)]+[(0-0)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(0-0)]+[(0-0)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(0-0)]+[(1-1)+(1-0)] = 2
 *
 *    .[3].(a).[3].(4).[3]. --> .[3].(8).[3].(2).[3].
 *          [s,i] + [s,j]        [(0-0)+(1-0)]+[(0-1)+(0-0)] = 0
 *          [i,t] + [t,j]        [(1-1)+(1-0)]+[(1-0)+(0-0)] = 2
 *          [i,u] + [j,u]        [(1-1)+(1-0)]+[(1-0)+(0-0)] = 2
 *
 *    .[9].(4).[9].(a).[9]. --> .[9].(2).[9].(8).[9].
 *          [s,i] + [s,j]        [(1-1)+(1-0)]+[(0-1)+(1-1)] = 0
 *          [i,t] + [t,j]        [(1-1)+(1-0)]+[(0-1)+(1-1)] = 0
 *          [i,u] + [j,u]        [(1-1)+(1-0)]+[(1-0)+(1-1)] = 2
 *
 *    .[9].(a).[9].(4).[9]. --> .[9].(8).[9].(2).[9].
 *          [s,i] + [s,j]        [(0-1)+(1-1)]+[(1-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-1)]+[(1-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-1)]+[(1-1)+(1-0)] = 2
 *
 *   if (i < j && a_i < MDO0  &&  a_i + a_j > MOD), then
 *     length((i,j)w) = length(w) - 1
 *       <=> (i < k < j  =>  (a_k < abs(a_i) || a_k > abs(a_j))
 *                j < k  =>  (abs(a_k) < abs(a_j) || abs(a_k) > abs(a_i))
 *
 *   if (i < j && a_j < MOD0  &&  a_i + a_j > MOD), then
 *     length((i,j)w) = length(w) - 1
 *       <=> (i < k      =>  (abs(a_k) < abs(a_i) || abs(a_k) >= abs(a_j))
 *
 *
 *  2.  s   i   t   j   u         s   i   t   j   u
 *    .[x].(8).[x].(a).[x]. --> .[x].(2).[x].(4).[x].
 *    .[x].(a).[x].(8).[x]. --> .[x].(4).[x].(2).[x].
 *    if x = 1, b, then the difference doesn't change
 *
 *    .[3].(8).[3].(a).[3]. --> .[3].(2).[3].(4).[3].
 *          [s,i] + [s,j]        [(0-1)+(0-0)]+[(0-0)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(0-0)]+[(0-0)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(0-0)]+[(1-1)+(1-0)] = 2
 *
 *    .[3].(a).[3].(8).[3]. --> .[3].(4).[3].(2).[3].
 *          [s,i] + [s,j]        [(0-0)+(1-0)]+[(0-1)+(0-0)] = 0
 *          [i,t] + [t,j]        [(1-1)+(1-0)]+[(0-1)+(0-0)] = 0
 *          [i,u] + [j,u]        [(1-1)+(1-0)]+[(1-0)+(0-0)] = 2
 *
 *    .[5].(8).[5].(a).[5]. --> .[5].(2).[5].(4).[5].
 *          [s,i] + [s,j]        [(0-1)+(1-0)]+[(0-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-0)]+[(1-0)+(1-0)] = 4
 *
 *    .[5].(a).[5].(8).[5]. --> .[5].(4).[5].(2).[5].
 *          [s,i] + [s,j]        [(0-1)+(1-0)]+[(0-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 4
 *
 *    .[7].(8).[7].(a).[7]. --> .[7].(2).[7].(4).[7].
 *          [s,i] + [s,j]        [(0-1)+(1-0)]+[(0-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-0)]+[(1-0)+(1-0)] = 4
 *
 *    .[7].(a).[7].(8).[7]. --> .[7].(4).[7].(2).[7].
 *          [s,i] + [s,j]        [(0-1)+(1-0)]+[(0-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-0)]+[(0-1)+(1-0)] = 4
 *
 *    .[9].(8).[9].(a).[9]. --> .[9].(2).[9].(4).[9].
 *          [s,i] + [s,j]        [(1-1)+(1-0)]+[(0-1)+(1-1)] = 0
 *          [i,t] + [t,j]        [(0-0)+(1-0)]+[(0-1)+(1-1)] = 0
 *          [i,u] + [j,u]        [(0-0)+(1-0)]+[(1-0)+(1-1)] = 2
 *
 *    .[9].(a).[9].(8).[9]. --> .[9].(4).[9].(2).[9].
 *          [s,i] + [s,j]        [(0-1)+(1-1)]+[(1-1)+(1-0)] = 0
 *          [i,t] + [t,j]        [(1-0)+(1-1)]+[(1-1)+(1-0)] = 2
 *          [i,u] + [j,u]        [(1-0)+(1-1)]+[(0-0)+(1-0)] = 2
 *
 *  if(i < j && a_i > MOD0 && a_j > MOD0 && TYPE_D), then
 *     length((i,-j)w) = length(w) - 1
 *        <=> (i < k < j => ((abs(a_k) < abs(a_i) && abs(a_k) < abs(a_j)))
 *                         ||(abs(a_i) < a_k < abs(a_j))||(a_i < a_k < a_j)
 *                 j < k => ((abs(a_k) < abs(a_i) && abs(a_k) < abs(a_j)))
 *
 */

int adj_br( UNIT *w , UNIT *s )
{
	int		i, j, k;
	UNIT	wu, wd, wa, *r;
	int		w_abs[MAX_rank], w_ord[MAX_rank+1];

	r = s;
	for( i = 0; i < siz0; i++ ){
		wu = w[i];
		wd = 0;
		for( j = i+1; j < rank; j++ ){
			if( w[j] > wu || w[j] < wd ) continue;
			*r++ = i+1;							/* (i,j)w is a required one */
			*r++ = j+1;
			wd = w[j];
		}
	}
	if( !wtype ) goto end;

	for( i = 0; i < rank; i++ )	w_ord[w_abs[i] = to_abs(w[i])] = i;

	for( i = 0; i < siz0; i++ ){
		wu = w[i];
		wa = w_abs[i];
		if( type_bc && w[i] > MOD0 ){
			for(j = i+1; j < rank && w_abs[j] < wa; j++);
			if( j >= rank ){
				*r++ = i+1;						/* (i)w is the required one */
				*r++ = MOD0;
			}
		}
		for( j = i + 1; j < rank; j++ ){
			if(wu + w[j] < MOD)
nextj:			continue;
			if(wu < MOD0){				/* a_i: positive,  a_j: negative */
				k = i+1;
chkabs:			for(; k < rank; k++){
					if(w_abs[k] > wu && w_abs[k] < w_abs[j])
						goto nextj;
				}
find:			*r++ = i+1;		/* (i,-j)w is a required one */
				*r++ = to_neg(j+1);
				continue;
			}else if(w[j] < MOD0){		/* a_i: negative,  a_j: positive */
				for(k = i+1; k < j; k++){
					if(w[k] > wu && w[k] < w_abs[j])
						goto nextj;
				}
				k = j+1;
				goto chkabs;
			}
			if(wtype == TYPE_D){		/* a_i: negative, a_j: negative */
				for(k = i+1; k < j; k++){
					if(w[k] < wa && w[k] < w_abs[j])
						continue;
					if(wa < w[k] && w[k] < w_abs[j])
						continue;
					if(wu < w[k] && w[k] < w[j])
						continue;
					goto nextj;
				}
				for(k = j + 1; k < rank; k++){
					if(w[k] < wa && w[k] < w_abs[j])
						continue;
					goto nextj;
				}
				goto find;
			}
		}
	}
	if( type_bc && is_neg(w[siz0]) ){			/* (rank)w is a required one */
		*r++ = rank;
		*r++ = MOD0;
	}
end:
	r[0] = 0;
	return( (int)(r-s) );
}

/*
 *	Positive roots which change negative by w
 *  In:  Weyl <w>
 *  Out: UNIT *s  (two bytes for every roots)
 *       ret  number of roots
 */
int chg_root( WEYL w, UNIT *s )
{
	UNIT	*r;
	int		i, j;

	r = s;
	for( i = 0; i < rank; i++ ){
		if( type_bc && is_neg(w[i]) ){
			*r++ = i+1;
			*r++ = MOD0;
		}
		for( j = i+1; j < rank; j++ ){
			if( w[i] > w[j] ){
				*r++ = i+1;
				*r++ = j+1;
			}
			if( w[i] + w[j] > MOD ){
				*r++ = i+1;
				*r++ = to_neg(j+1);
			}
		}
	}
	*r = 0;
	return( (int)(r-s)/2 );
}

static void found_r(int level)
{
	int		i;

	count++;
	for( i = level-1; i >= 0; i-- ){
		printf( "%d", r_exp[i] );
	}
	putchar( '\n' );
}

/*
 *	Get reduced expression for a given element
 *	mode =  0: get a reduced expression
 *	        1: get all reduced expression
 *          2: don't print
 *  return = length of w
 */
int get_r_exp(WEYL w0 , int mode)
{
	UNIT	u;
	int		i, level = 0, l = -1;
	AWEYL	w[246];

	for(i = 0; i < rank; i++)
		w[i] = w0[i];
start:
	if( level >= MAX_LEN ) return( ERROR );
	for( i = 1; i < rank; i++ ){
		if( w[i-1] > w[i] ){
			u = w[i];
			w[i] = w[i-1];
			w[i-1] = u;
			r_exp[level] = i;

check:		if( ++level != l ) goto start;
			found_r( level );
next:
			if( !level ) return( l );
			i = r_exp[--level];
			if( i < rank ){
				u = w[i-1];
				w[i-1] = w[i];
				w[i] = u;
				continue;
			}
			if( type_bc ) 	w[siz0] = to_neg(w[siz0]);
			else {
				u = w[siz0-1];
				w[siz0-1] = to_neg(w[siz0]);
				w[siz0] = to_neg(u);
			}
			goto next;
		}
	}
	if( wtype ){
	    if( type_bc ){
			if( w[siz0] > MOD0 ){					/* B_n or C_n */
				w[siz0] = to_neg(w[siz0]);
				r_exp[level] = rank;
				goto check;
			}
		}
		else if( w[siz0-1] + w[siz0] > MOD ){		/* D_n */
			u = w[siz0-1];
			w[siz0-1] = to_neg(w[siz0]);
			w[siz0] = to_neg(u);
			r_exp[level] = rank;
			goto check;
		}
	}
	if( l < 0 ){						/* get the first one */
		if( mode!=2 )
			found_r( level );
		if( mode!=1 ){
			r_exp[level] = 0;
			return( level );
		}
		l = level;
	}
	goto next;
}

/* --------------------------------------------------------------------
 *    Generate Young diagrams/Young tableau

    int next_yd(YD);
    int next_sum(UNIT *, int);
	int next_yt(YT, YD, int);
* -------------------------------------------------------------------- */

/*
 *         Generate all the partion (= Young diagram) of a fixed degree
 *
 *   In:  Young diagram <y>
 *   Out: ret > 0  # raws
 *                 <y> is changed to next one
 *              0 : END  (given <y> is the last one)
 */
int next_yd(YD y)
{
	int len, j, m, n;

	len = check_yd(y);
	if(len <= 0)
		return -2;
	j = len - 1;
	while(j >= 0 && y[j] == 1)
		j--;
	if(j < 0)
		return ERROR;
	m = len - j;
	n = --(y[j]);
	while(m)
		m -= (y[++j] = (m<n)?m:n);
	y[++j] = 0;
	return j;
}

/*
 *  Generate all the series of non-negative numbers whose sum and 
 *  length are fixed 
 *
 *  It comes from the inverse of the lexico grapic order
 *     300 210 201 120 111 102 030 021 012 003
 *
 *  In:  length <len>
 *       <s>  the former string of non-negative numbers
 *  Out:  ret 0:  <s> is changed into the next one
 *           -1:  Given <s> is the last one
 *     
 */
int next_sum(UNIT *s, int len)
{
	int i;

	for(i = len-1; ; ){
		if(--i < 0)
			return -1;
		if(s[i])
			break;
	}
	s[i]--;
	if(++i < len-1){
		s[i] += s[len-1] + 1;
		while(++i < len)
			s[i] = 0;
	}else
		s[i]++;
	return 0;
}

static int get_sumyd(UNIT *s, YD y)
{
	int i;

	*s++ = 0;
	*s++ = *y++;
	i = 1;
	while(*y){
		*s = *(s-1) + *y++;
		s++;
		i++;
	}
	*s = 0;
	return i;
}

/* 
 *                Generate all the Young Tableau in a lexicographic order
 *
 *  In:  Young Diagram <y>
 *       mode  0:  get the first one
 *             1:  get the next one
 *  Out: Young tableau <s>
 *
 *  1. Determine the maximal position which can be replaced to a larger number 
 *     without changing the numbers of the smaller positions
 *
 *  2. Rearrange the numbers in the boxes in the positions which are not 
 *     smaller than the above position.  It is easy to do so by starting from
 *    the smallest one in the numbers.
 *
 */
int next_yt(YT s, YD y, int mode)
{
	static UNIT ys[MAX_SIZE];	/* sum of raw length from the top raw */
//	static UNIT pos[MAX_SIZE];	/* position(0,..,rank-1) of the number(1,..,rank) */
	static int f_max_raw;	/* last raw */
	int i, j, k;
	int pt;					/* current position */
	int	num;				/* current number */
	int	raw;				/* current raw */
	int col;				/* current culumn */

#define	nm(m,n)	s[ys[(m)]+(n)]
// static int count;
//	if(count++ >= 10)
//		exit(1);
	if(mode == 0){
		f_max_raw = get_sumyd(ys, y);
		for(i=0; i<rank; i++){
			s[i] = i+1;
//			pos[i+1] = i;
		}
		s[rank] = 0;
		if(f_debug & FD_NEXTYT){
			printf("\nFirst Young Tableau: ");
			print_yt(y, s);
			printf("\n");
		}
		return 0;
	}else if(f_debug & FD_NEXTYT){
		printf("\nNext Young Tableau: ");
		print_yt(y, s);
		printf("\n");
	}
#if 1
# if 1
nxt:
	if(f_debug & FD_NEXTYT){
		for(i = 0; s[i]; i++)
			printf("%d", s[i]);
		printf(" ");
	}
	for(num = rank-2; num > 0; num--){		/* (0,0) = 1 */
								/* next element of the symmetric group */
		if(s[num] > s[num+1])
			continue;
		for(i = num + 2; i < rank && s[num] < s[i]; i++);
		k = s[--i];
		s[i] = s[num];
		s[num] = k;
#if 0
		if(k < s[num-1]){
			for(raw = f_max_raw-2; raw >= 0; raw--){
				if(ys[raw] == num)
					break;
				else if(ys[raw] < num)
					goto nxt;
			}
		}
#endif
		for(i = num + 1, j = rank - 1; i < j; i++, j--){
			k = s[i];
			s[i] = s[j];
			s[j] = k;
		}
		for(raw = f_max_raw - 1; raw >= 1; raw--){	/* order is correct? */
			for(i = y[raw]; --i > 0; ){
				if(nm(raw,i-1) > nm(raw,i))
					goto nxt;
				if(nm(raw-1, i) > nm(raw,i))
					goto nxt;
			}
			if(nm(raw-1,0) > nm(raw,0))
				goto nxt;
		}
		for(i = y[0]; --i > 0; ){
			if(s[i-1] > s[i])
				goto nxt;
		}
		if(f_debug & FD_NEXTYT){
			printf("->");
			print_yt(y,s);
			printf("\n");
		}
		return 0;
	}
# else
again:
	for(raw = f_max_raw -1; --raw >= 0; ){
		for(col = y[raw]; --col >=1 ; ){
			num = nm(raw,col);
			j = (y[raw+1] >= col)?(col-1):(y[raw+1]-1);
			for(i = raw+1; i < f_max_raw && y[i] > j; i++);
			--i;
			if(f_debug & FD_NEXTYT){
				printf("Number %d at (%d,%d) < %d at (%d,%d)?\n", 
					num, raw+1, col+1, nm(i,j), i+1, j+1);
			}
			if(num > nm(i,j))
				continue;
											/* (raw,col) can be larger */
			k = nm(i,j);
			pos = rs[i] + j;
			j = ys[raw] + col;
			for(i = j+1; i < rank; i++){
				if(s[i] > num && s[i] < k){
					pos = i;
					k = s[i];
				}
			}

			s[j] = k;			/* swap (raw, col) <-> pos */
			ys[k] = j;
			ys[num] = pos;

			for(i = ++j; i < rank; i++)
				s[i] = 0;
			for(num = 2; num <= rank; num++){
				if(pos[num] >= j){
					for(k = raw; k < f_max_raw; k++){
						j = ys[k+1] - 1;
						if(s[j])
							continue;	/* no empty box in this raw */
						j = ys[k];
						while(s[j])
							j++;
										/* the first empty box in this raw */
						col = ys[k] - j;
						if(col > 0 && s[j-1] > num)
							continue;
						if(raw > 0 && 
						  (nm(k-1,col) == 0 || nm(k-1, col) > num))
							continue;

						s[j] = num;		/* put num at (k,col) */
						pos[num] = j;
						if(f_debug & FD_NEXTYT)
							printf("%d -> (%d,%d)\n", num, raw+1, col+1);
						break;
					}
				}
			}
			return 0;
		}
	}
# endif
#else		/* The following is not correct */
	for(raw = f_max_raw -1, num = rank; --num >= 1; ){
		pt = pos[num];
		if(f_debug & FD_NEXTYT)
			printf("Number %dat%d ", num, pt+1);
		for(raw = f_max_raw; pt < ys[--raw]; );		/* (raw, col+1) = num */ 
		if(f_debug & FD_NEXTYT)
			printf("raw %d(%d->) ", raw+1, ys[raw]+1);
		col = pt - ys[raw] - 1;
		if(f_debug & FD_NEXTYT)
			printf("(%d, %d)[Legth of %d-th raw:%d %d]\n", raw+1, col+2, raw+2,
				 y[raw+1], col+1);
		if(y[++raw] <= col)						/* no box just below at col */
			col = y[raw]-1;
		for( ; ; raw++){
			if(raw >= f_max_raw)
				break;
			for(i = col; nm(raw, i) < num; i--);
			if(i < 0)
				continue;
			col = i;
			goto find;
		}
		continue;
find:	s[pos[num]] = rank;
		pt = ys[raw] + col;
		s[pt] = num;
		pos[num] = pt;
		if(f_debug & FD_NEXTYT)
			printf("put %d at (%d,%d)\n", num, raw, col);
		for(pt = 0, i = num + 1; pt < rank; pt++){
			if(s[pt] <= num)
				continue;
			s[pt] = i;
			pos[i++] = pt;
		}
		if(f_debug & FD_NEXTYT){
			printf("Get: ");
			print_yt(y,s);
			printf("\n");
		}
		return 0;
	}
#endif
end:
	if(f_debug & FD_NEXTYT)
		printf("The last one\n");
	return -1;
}

/* --------------------------------------------------------------------
 *
 * ------------------------------------------------------------------- */
void print_rs2w(YD yd, YT lef, YT rt)
{
	UNIT w[MAX_SIZE];

	rs2w(w, yd, lef, rt);
	set_rank(rank);
	print_wn(w, 1);
}

void print_tab(YD yd)
{
	int r;
	UNIT s[MAX_SIZE];

	r = next_yt(s, yd, 0);
	while(!r){
		print_yt(yd, s);
		printf(" ");
		r = next_yt(s, yd, 1);
	};
}

void printn(int num)
{
	if(is_neg(num)){
		if(f_mode & F_LONG){
			printf("-");
			num = to_abs(num);
		}else if(f_mode & F_OPTION){
			num = 9 + to_abs(num);
		}else
			num = to_pack(num);
	}
	printf("%c", num +((num < 10)?'0':('a'-10)));
}

void print_e(WEYL e)
{
	while(*e)
		printn(*e++);
}

void print_el(WEYL e, int len)
{
	while(len-- > 0)
		printn(*e++);
}

void print_yt(YD d, YT t)
{
	int i;

	i = *d++;
	if(!s_sep[0])
		printf("\\young(");
	while(*t){
		printn(*t++);
		if(--i <= 0 && *t){
			if(s_sep[0])
				printf(s_sep);
			else
				printf((*t)?",":")");
			i = *d++; 
		}
	}
	if(!s_sep[0])
		printf(")");

}

void print_ew(WEYL w, int len)
{
	int flag = 0;

	while(len-- > 0){
		if(flag++)
			printf(" ");
		printf("%d", *w++);
	}
}

void print_y(YD y)
{
	long dim;

	dim = dim_yd(y);
	if(dim < 1)
		err_exit(dim);
	else{
		print_e(y);
		printf("\t=%6ld\n", dim);
	}
}

void print_yd(int rank)
{
	int i, count;
	long dim, tmp;
	AWEYL y[MAX_SIZE];

	y[0] = rank;
	y[1] = 0;
	dim = count = 0;
	do{
		count++;
		for(i = 0; y[i]; i++)
			printn(y[i]);
		dim += (tmp = dim_yd(y));
		printf("\t=%6ld\n", tmp);
	}while(next_yd(y) > 0);
	printf("sum(%d)\t=%6ld\n\n", count, dim);
}


void print_wn(WEYL w, int f_short)
{
	int xx, yy, zz, vv;
	char *f;
	AWEYL y[MAX_SIZE]; // , ry[MAX_SIZE], ly[MAX_SIZE];

	f = (f_short)?"%x":"%3d";

	if(!f_short){					/* show the element */
		for(xx = 0; (yy = w[xx]) != 0; xx++)
			printf(f, yy);
	}else
		print_e(w);
	printf(" = ");
	inv_w(y, w);					/* show inverse */
	if(!f_short){
		for(xx = 0; (yy = y[xx]) != 0; xx++)
			printf(f, yy);
	}else
		print_e(y);
	printf("^{-1}");
	printf(f_short?" = ":"\n\n");
	get_r_exp(w, 2);
	vv = zz = strlen(r_exp);
	while(vv > 0 &&  (yy = r_exp[--vv]) != 0){
		if(!f_short)
			printf(f, yy);
		else
			printn(yy);
	}
	printf(" (%d)\n", zz);
}

/*
 *  In:  element of Weyl group
 *  Out: print the element
 */
void print_w(WEYL w)
{
	int xx, yy, zz, ww, vv, f_short;
	char *f, *r;
	AWEYL y[MAX_SIZE], ry[MAX_SIZE], ly[MAX_SIZE];

	if(strlen(w) > 16){
		f = "%3d";
		f_short = 0;
	}else{
		f = "%x";
		f_short = 1;
	}
	r = (f_mode & F_SHORT)?(s_sep[0]?s_sep:","):"\n";
	print_wn(w, f_short);
	if(wtype != TYPE_A)
		return;
    rs(w, y, ry, ly);
	ww = y[0];
	if((f_mode & F_SHORT) && !s_sep[0])
		printf("\\young(");
	for(xx = zz = 0; (yy = y[xx])!= 0; ){
		if(xx++)
			printf(r);
		for(vv = yy; vv--; )
			printf(f, ly[zz++]);
		if(f_short && !(f_mode & F_SHORT)){
			zz -= yy;
			for(vv = ww - yy + 2; vv--; )
				printf(" ");
			for(vv = yy; vv--; )
				printf(f, ry[zz++]);
		}
	}
	if(!f_short)
		printf("\n");
	if(!f_short || f_mode & F_SHORT){
		printf((f_mode & F_SHORT)?(s_sep[0]?"\t":")\t\\young("):"\n");
		for(xx = zz = 0; (yy = y[xx])!= 0; ){
			if(xx++)
				printf(r);
			while(yy--)
				printf(f, ry[zz++]);
		}
	}
	if((f_mode & F_SHORT) && !s_sep[0])
		printf(")");
	printf("\n");
}

/*
 *   In:  Young diagram
 *   Out: print all the elements in the two sided cell corresponding to the diagram
 */
void print_2cell(YD y)
{
	int i, j, k, pt, pt2, d, r, len;
	AWEYL *yt, tab1[MAX_SIZE], tab2[MAX_SIZE], ww[MAX_SIZE];

	len = check_yd(y);
	if(y < 0)
		return;
	if(f_mode & F_ORD)
		w_count = 0;
	for(rank = i = 0; i < len; i++)
		rank += y[i];
	set_rank(rank);
	d = dim_yd(y);
	if(d <= 0 || d >= 10000)
		return;
	yt = (AWEYL *)malloc(d * rank);
	if(yt == NULL)
		return;
	for(i = pt = 0, r = next_yt(tab1, y, 0); !r && i < d; i++, r = next_yt(tab1, y, 1)){
		for(k = 0; k < rank; )
			yt[pt++] = tab1[k++];
	}
	if(i < d)
		d = i;
	for(i = pt = 0; i < d; i++){
		for(k = 0; k < rank; )
			tab1[k++] = yt[pt++];
		tab1[k] = 0;
		for(j = pt2 = 0; j < d; j++){
			for(k = 0; k < rank; )
				tab2[k++] = yt[pt2++];
			tab2[k] = 0;
			rs2w(ww, y, tab1, tab2);
			if(f_mode & F_ORD)
				add_w_buf(ww);
			else{
				if(f_mode & F_ROOT){
					printf("\n");
					print_yt(y, tab1);
					printf( " ");
					print_yt(y, tab2);
					printf( " ");
				}
				if(!(f_mode & F_LONG)){
					print_e(ww);
					printf(" ");
				}else{
					print_w(ww);
					printf("\n");
				}
			}
		}
	}
	free(yt);
	if(f_mode & F_ORD)
		print_w_buf(0);
}

/*
 *  In:  Young tableau
 *  Out: print all the elements of the left/right cell corresponding to the Young tableau
 */
void print_cell(YD y, YT tab)
{
	int r;
	AWEYL tab2[MAX_SIZE], ww[MAX_SIZE];

	if(f_mode & F_ORD)
		w_count = 0;
	for(r = next_yt(tab2, y, 0); !r; r = next_yt(tab2, y, 1)){
	    if(f_mode & F_NEG)
			rs2w(ww, y, tab2, tab);
		else
			rs2w(ww, y, tab, tab2);
		if(f_mode & F_ORD)
			add_w_buf(ww);
		else if(!(f_mode & F_LONG)){
			print_e(ww);
			printf(" ");
		}else{
			print_w(ww);
			printf("\n");
		}
	}
	if(f_mode & F_ORD)
		print_w_buf(0);
}

/*
 *   print Young tableau in a short format 
 */
void sprint_yt(UNIT *buf, YD y, YT tab)
{
	int i, n, f;
#ifdef	NOW
for(i=0;i < 6; i++)
	printf("(%d)", y[i]);
for(i=0;i < 7; i++)
	printf("%d", tab[i]);
printf("\n");
#endif

	f = 0;
	i = *y++;
//	if(!s_sep[0]){
//		strcpy(buf, "\\young(");
//		buf += 7;
//	}
	while(*tab){
		n = *tab++;
		*buf++ = (n<10)?(n+'0'):(n+('a'-10));
		if(--i <= 0 && *tab){
			f++;
			*buf++ = //(s_sep[0])?s_sep[0]:',';
				s_sep[0];
			i = *y++; 
		}
	}
	if(!f && s_sep[0] != ' ')
		*buf++ = s_sep[0];
//	if(!s_sep[0])
//		*buf++ = ')';
	*buf = 0;
}

/*
 *  Add  Young tableau <rs> to UNIT *w_buf
 *  ret:  >= 0  already exists
 *          -1  add it
 */
int add_yt(UNIT *rs)
{
	int	i, len;

	for(i = 0; i < w_count; ){
		if(!strcmp(w_buf + i, rs))		/* same tableau? */
			return i;		
		while(w_buf[i++]);				/* skip to the next */
	}

	len = strlen(rs) + 1;				/* new tableau */
	if(w_count + len >= MAX_BUFW)		/* buffer Overflow */
		err_exit(ER_TBUFOVER);
	strcpy(w_buf + w_count, rs);
	w_count += len;
	return -1;
}

/*
 *   Get adjacent left cells
 */
void adj_cell(YD y, YT tab, int mode)
{
	int i, f_tmp, w_num, n, tmp;
	UNIT *ww, *wp;
	UNIT w[MAX_SIZE], r[MAX_SIZE*2], ro[MAX_SIZE*2], ly[MAX_SIZE], ry[MAX_SIZE], yy[MAX_SIZE];

	f_tmp = f_mode;
	f_mode |= F_NOPRINT|F_ORD|F_NEG;
	print_cell(y, tab);					/* get elements in the right cell */
	w_num = w_count;
	w_count = 0;
	ww = (UNIT *)malloc(rank*w_num);
	memmove(ww, w_buf, rank*w_num);		/* reserve the elements */
	sprint_yt(r, y, tab);				/* set the original cell as a string */
	add_yt(r);
	strcpy(ro, r);

	for(n = 0; n < w_num; n++){
		wp = ww + rank*n;				/* get the <n>-the element in the cell */
		for(i = 0; i < rank; i++)
			w[i] = wp[i];
		w[i] = 0;

		for(i = 1; i < rank; i++){
			if(w[i-1] < w[i]){			/* to larger element */
				if(mode)
					continue;
			}else if(!mode)				/* to smaller element */
				continue;
			tmp = w[i-1];
			w[i-1] = w[i];
			w[i] = tmp;
			rs(w, yy, ly, ry);			/* get the RS correspondence */
			sprint_yt(r, yy, ry);		/* get the right cell as a string */
			if(add_yt(r) < 0){			/* reseve it in the buffer */
				if(s_sep[0] == ' '){
					printf("(%s)>(%s)\n", ro, r);
				}else{
					printf(r);				/* It is a new one */
					printf(" ");
				}
			}
			tmp = w[i-1];				/* recover the original element */
			w[i-1] = w[i];
			w[i] = tmp;
		}
	}
	free(ww);
	f_mode = f_tmp;
}

/*
 *   In:  An element of symmetric group
 *   Out: print all the elements of the left/right/two sided cell belonging 
 *        to the given element
 */
void print_wcell(WEYL w)
{
	AWEYL y[MAX_SIZE], ry[MAX_SIZE], ly[MAX_SIZE];

	rs(w, y, ly, ry);
	if(f_mode & F_2CELL)
		print_2cell(y);
	else
        print_cell(y, (f_mode & F_NEG)?ry:ly);
}

#if 0
/*
 *  rank is given
 *  All the element of the Weyl group is given with some information
 *    inverse, reduced expression, length and RS correspondence (TYPE_A)
 */
void print_wy(int r)
{
	int i;
	PACK pk, last;
	AWEYL w[MAX_SIZE], *pt;

	for(last = i = 1; i <= rank; i++)
		last *= i;
	if(wtype != TYPE_A){
		for(i = (wtype == TYPE_D)?1:0; i < r; i++)
			last *= 2;
	}
	if(last < w_max){
		for(pt = w_buf, pk = 0; pk < last; pt += r)
		unpack(pt, pk++, r);
		w_count = last;
		sort_by_length();
		w[r] = 0;
		for(pt = w_buf, pk = 0; pk < last; pk++, pt += r){
			for(i = 0; i < r; i++)
				w[i] = pt[i];
				print_w(w);
				if(wtype == TYPE_A)
					printf("\n");
		}
	}
	else for(pk = 0; pk < last; ){
		unpack(w, pk++, r);
		print_w(w);
		if(wtype == TYPE_A)
			printf("\n");
	}
}
#endif

/*
 *	print roots from data UNIT* <r>  which is null terminated
 */
void print_r(UNIT *r)
{
	int m, n;

	while((m = *r++) != 0){
		n = *r++;
		printf("%c", (n<=MOD0)?'(':'[');
		printn(m);
		m = n;
		if(m != MOD0){
			if(is_neg(n))
				m = to_neg(m);
			printn(m);
		}
		printf("%c", (n<=MOD0)?')':']');
	}
}

/*
 *  all the roots changed to negatives by a given element of Weyl group are shown
 */
void print_root(WEYL w)
{
	chg_root(w, w_buf);
	print_r(w_buf);
}

/*
 *  show the order of two elements of Weyl group wrt to Bruhat ordering
 */
void print_cmp(WEYL w1, WEYL w2)
{
	int r;
	char *msg;

	r = comp_b(w1, w2);
	switch(r){
		case 0:
			msg = "=";
			break;
		case 1:
			msg = ">";
			break;
		case -1:
			msg = "<";
			break;
		case 2:
			msg = "<>";
			break;
		default:
			msg = "?";
			break;
	}
	print_e(w1);
	printf(" %s ", msg);
	print_e(w2);
	printf("\n");
}

/*
 *  Initialize constants related to rank
 */
void set_rank(int r)
{
	rank = r;
	rank2 = 2*rank + 1;
	siz0 = rank - 1;
	siz1 = (wtype==0)? (rank-1):rank;
	w_max = MAX_BUFW/rank - 1;
}

/* -------------------------------------------------------------------------
 *                                Messages
 * ------------------------------------------------------------------------- */
static char *msgo = 
	"  << Robinson-Schensted, Young Tableau, Bruhat order, Weyl/Linear group >>\n"
	"\t   Ver.0.5   Written by Toshio OSHIMA(July 25, 1999/2005)\n\n"
	"Usage: rs [-?[DWYL]] [-WYL] [-ABCD] [-abcslo] [-d<num>] [-S<sep>] <exp> [<exp>]\n\n"
    " -?[DWYL]: show help for Dynkin diagram, -W, -Y, -L option\n"
    "    ex.  rs -?W\n"
	" -W: Weyl group (default)\n"
	"     inverse, reduced expression, length, RS correspondence, double cosets\n"
	"     Sign change of n is expressed in <exp> by  2*rank+1-n  or (-n) or\n"
	"     a=-1, b=-2,... with -o option (rank<10)\n"
	"   ex.  rs [-W] 34512   (a=10, b=11,...,z=35 in general)\n"
	"     rs -Ba 462 is same as  rs -Ba (-3)-12  or  rs -Bao ca2\n"
	" -Y: Young Tableau (<exp>:tableau/rank)\n"
	"     dimension of rep. of Weyl group.      ex.  rs -Y 4221\n"
	" -L: Linear group (<exp>:weight/sum of coeff.)\n"
	" -ABCD: Type of the group (default:A)\n"
	" -d: <num> is the rank/degree (=number)    ex.  rs [-W|-Y] -d5\n"
	" -s: short expression of Young Tableau (default: -S/  TeX: -S)\n"
	" -l: '-' sign expression     -o: another expression (a=-1,...)\n"
	" -a: all reduced expression\n"
	" -b: Bruhat order\t-c: cells\t-r: roots\n"
;
static char *msgw = 
	"\t\t<< -W option is for Weyl group >>\n"
	"      show inverse, reduced expression, length, RS correspondence(default)\n"
	"      If two elements are given, compare them by Bruhat order(without -*)\n" 
	"-ABCD: Type (default A)\n"
	"<exp> indicates the element (permutation with changing sign)\n"
	"     Negative number -<n> is expressed by  2*rank+1-<n> or (-<n>)\n"
	"      or a=-1, b=-2,... with -o option\n"
	"     V<exp>: Inverse of <exp>,  V2341 means 4123\n"
	"     P<exp>: the longest element of the parabolic corresponding to <exp>\n"
	"       P23, Q23 and -Bo P23 mean 21543, 34512 and 21cd, respectively\n"
	"     Q<exp>: P<exp> * longest, R<exp>: inverse of P<exp>, S<exp>: simple refl.\n"
	"     <exp> is the rank under the -d option\n"
	"-*   multiply two <exp>'s.  -* 1342 2134 means 3142\n"
	"-r   show roots changed from positive to negative by the element\n"
	"-b   show bigger adjacent elements under Bruhat order\n"
	"    -r: corresponding reflections from right\t--: bigger -> smaller\n"
	"    -d<rank>: Bruhat poset of the <rank>   ex. rs -Bbd3o\n"
	"      <n>:<x><y>.. means the upper <n>-th element < the lower <x>,<y>..-th\n" 
	"    -b0: order under the simple reflections from right (-b1: from left)\n" 
	"-c   show elements belonging to the same left cell\n"
	"    --: right cell  -c2[r]: two sided cell[with tableau]  -b: with Bruhat order\n"
	"-w[-[b]] <exp> <exp>:  double coset.  <exp> is same as <exp> in P<exp>\n"
	"     show shortest[longest] representatives [with Bruhat relation]\n"
	"-p[-] <exp> <exp>: parabolic decomposition. <exp> are element and parabolic"
;

static char *msgy = 
	"\t\t<< -Y option is for Young diagram/tableau >>\n\n"
	"<exp> indicates a Young diagram (decreasing numbers ex. 431) or -Yod<num>\n"
	"<num> is the degree of Young diagrams under -od option\n"
	"      show the dimension of the representation of the symmetric group ([-S])\n"
	"   -c[r|b]: two sided cell [with tableau/Bruhat order]\n\n"
	"<exp> indicates a Young tableau (ex, 1346/25/7/) or -Yd<num>\n"
	"<num> is the degree of Young tableaus under -od option\n"
	"      show tableau with the same shape ([-S])\n"
	"   -c[b]: show elements in the left cell given by the tableau\n"
	"      [with Bruhat order, -b0 -b1]\n" 
	"      --: left cell -> right cell\n"
	"   -s show the larger adjacent left(right) cells\n"
	"      --: larger -> smaller\n"
	"\n"
	"<exp> <exp> two tableau give the inverse of Robinson-Schensted correspondence\n"
	"      ex.  rs -Y 1346/25/7 1347/26/5\n"
	"\n" 
	" ####  4     1346\n"
	" ###   3     25\n"
	" #     1     7\n"
;

static char *msgl = 
	"\t\t<< -L option is for linear group >>\n\n"
	"      show the dimension of the representation of the linear group\n\n"
	"-ABCDEFG: Type (default A)\n"
	"<exp> indicates the coefficients (<128) of the fundamental weights or \n"
	"-d<num> <exp> represents the rank <num> and sum <exp> of coeff.\n"
	"  ex.  rs -LBd5 4     rs -LF 1243     rs -L (12)34\n\n"
	" o-o-...-o   o-o-...-o=>o   o-o-...-o<=o   o-o-...-o-o\n"
	" 1 2     n   1 2    n-1 n   1 2    n-1 n   1 2     | n-1\n"
	"                                                   o n\n\n"
	" o-o-o-o-o  o-o-o-o-o-o  o-o-o-o-o-o-o   o-o=>o-o   o<Eo\n"
	" 1 3 | 5 6  1 3 | 5 6 7  1 3 | 5 6 7 8   1 2  3 4   1  2\n"
	"     o 2        o 2          o 2\n"
;

static char *msgd = "\n"
	" -?D[A][B][C][D][F][E6][E7][E8][G][R] for more information\n"
    "   with M, K and dimensions of the root decompositions and symmetric spaces\n\n"
    " 1 1     1   1  2      1 2     2  2      2 2     2  1   1 2     2 1\n"
	" o-o-...-o   o=>o<=c   o-o-...-o=>o   c=>o-o-...-o<=o   o-o-...-o-o\n"
	" |..-c-..|               |                                |     |\n"
	"                         c                                c     o1\n\n"

	" 1 2 3 2 1       2 3 4 3 2 1     2 4 6 5 4 3 2         2 3  4 2   3  2\n"
	" o-o-o-o-o     c-o-o-o-o-o-o     o-o-o-o-o-o-o-c     c-o-o=>o-o   o<Eo-c\n"
	"     |               |               |\n"
	"     o2              o2              o3\n"
	"     |\n"
	"     c\n"
    " 1,4,5,7,8,11 1,5,7,9,11,13,17 1,7,11,13,17,19,23,29  1,5,7,11    1,5\n"
;

static char *msgdr = "\n"
	"Simple roots:\n"
    "A(n)'s j-th root: e_j-e_{j+1}    B(n)'s n-th root: e_n\n"
    "E(n) 1: (e_1+e_8)/2-(e_2+e_3+...+e_7)/4    2: e_2+e_1\n"
    "     j: e_{j+1}-e_j  (j=3,4,...)\n"
    "F(4)  : e_2-e_3  e_3-e_4  e_4  (e_1-e_2-e_3-e_4)/2\n"
    "G(2)  : e_1-e_2  -2e_1+e_2+e_3\n\n"
    "Value of 2 rho and  <rho,rho>:\n"
    "A(n) : sum (n-2(j-1))e_j\t\t\tn(n+1)(n+2)/12\n"
    "D(n) : sum 2(n-j)e_j\t\t\t\tn(n-1)(2n-1)/6\n"
    "BC(n;r,m,s): sum (2m(n-j)+r+2s)e_j\n"
    "\t\tm^2n(n-1)(2n-1)/6+mn(n-1)t+nt^2  with  t=(r+2s)/2\n"
    "E(6): 2e_2+4e_3+6e_4+8e_5- 8e_6- 8e_7+ 8e_8\t78\n"
    "E(7): 2e_2+4e_3+6e_4+8e_5+10e_6-17e_7+17e_8\t399/2\n"
    "E(8): 2e_2+4e_3+6e_4+8e_5+10e_6+12e_7+46e_8\t620\n"
    "F(4;k,m): (5k+6m)e_1+(k+4m)e_2+(k+2m)e_3+ke_4\t7k^2+18km+14m^2\n"
    "G(2): -2e_1-4e_2+6e_3\t\t\t\t14\n"
;

static char *msgla = "\n"
	"SL(n+1,R)    A(n:1)           n*2          o-o-o-o-..-o-o        SO(n)\n"
    "   (n+1)^2-1 = 2xn(n+1)/2   + n + 0                              n(n+3)/2\n\n"
    "SL(n+1,C)    A(n;2)        (n-1)SO(2)                            SU(n)\n"
    "  2(n+1)^2-2 = 2xn(n+1)     + n + n                              (n+1)^2-1\n\n"
	"SU*(2(n+1))  A(n:4)        (n+1)SU(2)      *-o-*-o-..-o-*        Sp(n+1)\n"
    "  4(n+1)^2-1 = 2x2n(n+1)    + n + 3(n+1)                         n(2n+3)/2\n\n"
    "SU(n+m,n)    BC(n:2m,2,1)  SU(m)xnSU(1)    o-..-o-*-..-*-o-..-o\n"
    "  (2n+m)^2-1 = 2xn(2n+2m-1) + n + m^2-1+n  |----|--------|----|  2n(n+m)\n\n"
    "SU(n,n)      C(n:2,1)      (n-1)SU(1)x2    o-o-..-o-o-o-..-o\n"
    "   4n^2-1    = 2xn(2n-1)    + n + n-1      |-|-..-|---|-..-|     2n^2\n"
;

static char *msglb = "\n"
	"SO(n+2m+1,n) B(n:2m+1, 1) SO(2m+1)x(n-1)*2 o-o-..-o-*-*-..-*=>*\n"
    " (m+n)(2n+2m+1) = 2xn(n+2m) + n + m(2m+1)                        n(2n+2m+1)\n\n"
    "SO(n+1,n)    B(n:1,1)          n*2         o-o-..-o-o-o-..-o=>o  \n"
    "   n(2n+1)      = 2xn^2     + n + 0                              n(2n+1)\n\n"
    "SO(2n+1,C)   B(n:2,2)         nSO(2)                             SO(2n+1)\n"
    "  2n(2n+1)      = 2x2n^2    + n + n                              n(2n+1)\n"
;

static char *msglc = "\n"
	"Sp(n,R)      C(n:1,1)         n*2        o-o-o...-o-o<=o         U(n)\n"
    "  n(2n+1)        = 2xn^2        + n + 0                          n(n+1)\n\n"
	"Sp(n,C)      C(n:2,2)        nSO(2)                              Sp(n)\n"
    " 2n(2n+1)        = 2x2n^2       + n + n                          n(2n+1)\n\n"
	"Sp(n+m,n)    BC(n:4m,4,3)  Sp(m)xnSU(2)  *-o-*-..-o-*-..-*<=*    Sp(n+m)xSp(n)\n"
    " (2n+m)(4n+2m+1) = 2xn(4n+4m-1) + n + (m(2m+1)+3n)               4n(n+m)\n\n"
    "Sp(n,n)      C(n:4,3)       nSU(2)       *-o-*-..-o-*-o-*<=o     Sp(n)xSp(n)\n"
    "  2n(4n+1)       = 2xn(4n-1)    + n + 3n                         4n^2\n\n"
;

static char *msgld = "\n"
	"SO(n+2m,n)   BC(n:2m,1)   SO(2m)x(n-1)*2  o-o-..-o-*-..-*-*-*\n"
    " (m+n)(2n+2m-1) = 2xn(n+2m-1) + n + m(2m-1)               |       n(m+n)\n"
    "                                                          *\n"
    "SO(n+2,n)     B(n:2,1)    SO(2)x(n-2)*2   o-o-..-o-o-..-o-o-o\n"
    "   (n+1)(2n+1)  = 2xn(n+1)    + n  + 1                    |/      n(n+2)\n"
    "                                                          o\n"
    "SO(n,n)       D(n:1)          n*2         o-o-..-o-o-..-o-o-o\n"
    "    n(2n-1)     = 2xn(n-1)    + n  + 0                    |       n^2\n"
    "                                                          o\n"
    "SO(2n,C)      D(n:2)         nSO(2)                               SO(2n)\n"
    "    2n(2n-1)    = 2x2n(n-1)   + n  + n                            n(2n-1)\n\n"
    "SO*(4n)      C(n:4,1)        nSU(2)x2     *-o-..-*-o-..-*-o-*     U(2n)\n"
    "    2n(4n-1)    = 2xn(4n-3)   + n  + 3n                   |       2n(2n-1)\n"
    "                                                          o\n"
    "SO*(4n+2)    BC(n:4,4,1)    nSU(2)xSO(2)  *-o-..-*-o-..-o-*-o     U(2n+1)\n"
    "  (2n+1)(4n+1)  = 2xn(4n+1)   + n  + 3n+1                 |/      2n(2n+1)\n"
    "                                                          o\n"
;

static char *msgle6 = "\n"
    "                                           |-|---|-|\n"
	"EI          E(6:1)             6*2         o-o-o-o-o    Sp(4)/2\n"
    "     78   = 2x36  + 6 +  0                     |        36 + 42\n"
    "                                               o\n"
	"E6          E(6:2)           6SO(2)                     E6\n"
    "    156   = 2x72  + 6 +  6                              78 + 78\n\n"
    "EII        F(4:2,1)         2SO(2)x2*2     o-o-o-o-o    SU(6)xSU(2)/2\n"
    "     78   = 2x36  + 4 +  2                     |        38 + 40\n"
    "                                               o\n"
    "                                           |-|---|-|\n"
    "EIII       BC(2:8,6,1)      SU(4)xSO(2)    o-*-*-*-o    Spin(10)xSO(2)/2\n"
    "     78   = 2x30  + 2 + 16                     |        32 + 46\n"
    "                                               o\n"
    "EIV        A(2:8)             Spin(8)      o-*-*-*-o    F4\n"
    "     78   = 2x24  + 2 + 28                     |        26 + 52\n"
    "                                               *\n"
;

static char *msgle7 = "\n"
	"EV         E(7:1)            7*2         o-o-o-o-o-o    SU(8)/2\n"
    "    133   = 2x63  + 7 +  0                   |          63 + 70\n"
    "                                             o\n"
	"E7         E(7:2)           7SO(2)                      E7\n"
    "    266   = 2x126 + 7 +  7                              133 + 133\n\n"
    "EVI        F(4:4,1)        3SU(2)x2*2    o-o-o-*-o-*    Spin(12)xSU(2)/2\n"
    "    133   = 2x60  + 4 +  9                   |          64 +  69\n"
    "                                             o\n"
    "EVII       C(3:8,1)        SO(8)x2       o-*-*-*-o-o    E6xSO(2)/2\n"
    "    133   = 2x51  + 3 + 28                   |          54 +  79\n"
    "                                             *\n"
    "EVIII      E(8:1)            8*2         o-o-o-o-o-o-o  Spin(16)/2\n"
    "    248   = 2x120 + 8 +  0                   |          120 + 128\n"
    "                                             o\n"
    "E8         E(8:2)          8SO(2)                       E8\n"
    "    248   = 2x240 + 8 +  8                              248 + 248\n\n"
    "EIX        F(4:8,1)        SO(8)x2*2     o-*-*-*-o-o-o  E7xSU(2)/2\n"
    "    248   = 2x108 + 4 + 28                   |          112 + 136\n"
    "                                             *\n"
;

static char *msglf = "\n"
	"FI        F(4:1,1)            4*2         o-o=>o-o      Sp(3)xSU(2)/2\n"
    "     52   = 2x24   + 4 +  0                             24 + 28\n\n"
	"F4        F(4:2,2)          4SO(2)                      F4\n"
    "    124   = 2x48   + 4 +  4                             52 + 52\n\n"
    "FII       BC(1:8,7)         Spin(7)       *-*=>*-o      Spin(9)\n"
    "     52   = 2x15   + 1 + 21                             16 + 36\n\n"
    "G         G(2:1)              2*2           o<Eo        SO(4)\n"
    "     14   = 2x6    + 2 +  0                             6  +  8\n\n"
    "G2        G(2:2)            2SO(2)                      G2\n"
    "     28   = 2x12   + 2 +  2                             14 + 14\n"
;

static void msg(void)
{
	fprintf(stderr, msgo);
}

 /*
 *  TYPE_B/TYPE_D : multiply by 2 for Spin representation
 */
void coef2w(WEYL w, int r)
{
	int i;

	set_rank((wtype==TYPE_A)?(r+1):r);

	if(wtype == TYPE_B || wtype== TYPE_D){
		if(wtype == TYPE_D)
			w[r-1] -= w[r-2];
		for(i = r - 2; i >= 0; i--)
			w[i] *=2;
	}
	if(f_debug & FD_FUNDWEIGHT){
		printf("Fundamental weight: ");
		for(i = 0; i < rank; i++)
			printf("%d ", w[i]);
		printf("\n");
	}
	for(i = r-2; i >= 0; --i)
		w[i] += w[i+1];
}

/*
 *  rho shift of weight
 */
void add_rho(WEYL w)
{
	int i, n;

	if(f_debug & FD_RHOSHIFT){
		printf("Weight: ");
		for(i = 0; i < rank; i++)
			printf("%d ", w[i]);
		printf("\n");
	}

	if(wtype == TYPE_A || wtype == TYPE_C){
		for(i = 0; i < rank; i++)
			w[i] += rank - i;
	}else{
		n = 2*rank;
		if(wtype == TYPE_B)
			n++;
		for(i = 0; i < rank; i++){
			n -= 2;
			w[i] += n;
		}
	}
}

/* --------------------------------------------------------------------
 *				Weyl's dimension formula
 * -------------------------------------------------------------------- */
static char *D_E6[] =
{
	"012101", "014-1-030125", ""
};

static char *D_E7[] = {
	"0121111", "340-4-31342-4643210", 
	"1233211", "",
	""
};

static char *D_E8[] = {
	"01211111", "3450-5-4-313452-5-4743-7547-4-3",
	"12432112", "15-14310-3-4-5",
	"13543212", "01",
	"13543213", "0123456",
	"13433212", "",
	""
};

static char *D_F[] =
{
	"0120", "33012210", 
	"1120", "132", 
	"1121", "",
	""
};

static char *D_G[] =
{
	"21", "001",""
};

static void dim_sub(WEYL w, char **d)
{
	int	j, n, c, m;
	char *s;

	for( ;*(s = *d++); ){
		for(j = n = c = 0; s[j]; j++){
			m = s[j] - '0';
			c += m;
			n += m*w[j];
		}
		power[n + c]++;
		power[c]--;

		if(!*(s = *d++))
			break;
		for(j = 0; s[j]; ){
			m = s[j++];
			if(m == '-'){
				n -= w[s[j++] - '0'];
				c--;
			}else{
				n += w[m - '0'];
				c++;
			}
			power[n + c]++;
			power[c]--;
		}
	}
}

/*
 *   Weyl's dimension formula for all the simple Lie groups
 *   In:  the coefficients of fundamenal Weights
 */
long dim_rep2(WEYL w)
{
	int i, j, n, n0, d, d0, tmp, ra;
	UNIT ws[8];
	char **pt;
	char *s;

	for	(i = n = 0; i < rank; i++)
		n += w[i];

	ra =rank;
	if(wtype == TYPE_A)
		max_power = n+rank;
	else if(wtype == TYPE_F){
		if(rank != 4){
			err_exit(ER_BADRANK);
			return ERROR;
		}
		max_power = 8*(n+rank); 
	}else if(wtype == TYPE_G){
		if(rank != 2){
			err_exit(ER_BADRANK);
			return ERROR;
		}
		max_power = 3*(n+rank);
	}else{
		if(wtype == TYPE_E){
			if(rank < 6 || rank > 8){
				err_exit(ER_BADRANK);
				return ERROR;
			}
			tmp = w[1];
			for(i = 2; i < rank ; i++)
				w[i-1] = w[i];
			w[rank-1] = tmp;
			max_power = (ra-3)*(n+rank);
		}else
			max_power = 4*(n+rank);
	}
	if(max_power >= MAX_POWER)
		return ERROR;
	clear_int(power, max_power);

	if(wtype != TYPE_A)
		ra--;
	if(wtype != TYPE_F && wtype != TYPE_G){
		for(i = 0; i < ra; i++){			/* TYPE_A  0001111000  e_i-e_j */
			for(n = 0, j = i+1; j <= ra; j++){
				n += w[j-1];
				power[n + j - i]++;
				power[j - i]--;
			}
		}
	}

	switch(wtype){
	  case	TYPE_B:
		for(n = w[ra], d = 1, i = ra; ; ){			/* 00001111 */
			power[n+d]++;
			power[d]--;
			if(--i < 0)
				break;
			n += 2*w[i];
			d += 2;
		}
		for(n0 = w[ra], d0 = 1,  i = ra-1; i >= 0; i--){	/* 00011222 */
			n0 += w[i];
			d0++;
			for(n = n0, d = d0, j = ra-1; ; j--){
				power[n+d]++;
				power[d]--;
				if(j <= i)
					break;
				n += w[j];
				d++;
			}
		}
		break;

	  case TYPE_C:
		for(n0 = 2*w[ra], d0 = 2, i = ra; ;){	/* 0011221 e_i+e_j, 2e_i */
			for(n = n0, d = d0, j = ra; ;){
				power[n+d]++;
				power[d]--;
				if(j <= i)
					break;
				n += w[--j];
				d++;
			}
			if(i-- <= 0)
				break;
			n0 += w[i];
			d0++;
		}
		break;

	  case TYPE_D:								/* 0001111101 */
		power[w[ra]+1]++;
		for(n = w[ra], i = ra-2; i >= 0; i--){
			n += w[i];
			power[n + ra - i]++;
			power[ra - i]--;
		}										/* 0001112211 */
		for(n0 = w[ra] + w[ra-1], i = ra-2; i >= 0; i--){
				n0 += w[i];
			for(n = n0, j = ra-1; ;){
				power[n + 2*ra - i - j]++;
				power[2*ra - i - j]--;
				if(--j <= i)
					break;
				n += w[j];
			}
		}
		break;

	  case TYPE_E:
		power[w[ra]+1]++;
		for(n = w[ra], i = 2; i < ra; i++){		/* mutiplicity <= 0 */
			n += w[i];
			power[n + i]++;
			power[i]--;
			power[n + i + w[1] + 1]++;
			power[i+1]--;
			power[n + i + w[1] + w[0] + 2]++;
			power[i+2]--;
		}
		for(i = 0; i < 5; i++)			/* E6 */
			ws[i] = w[i];
		ws[5] = w[ra];
		dim_sub(ws, D_E6);
		if(rank > 6){
			ws[5] = w[5];				/* E7 */
			ws[6] = w[ra];
			dim_sub(ws, D_E7);
			if(rank == 8){				/* E8 */
				ws[6] = w[6];
				ws[7] = w[ra];
				dim_sub(ws, D_E8);
			}
		}
		tmp = w[ra];
		for(i = ra; i >= 2; i--)
			w[i] = w[i-1];
		w[1] = tmp;
		break;

	  case TYPE_F:
		for(i = 0; i < rank; i++){		/* multiplicity <= 1 */
			if(i < 2){
				n = 2*w[i];
				d = 2;
			}else{
				n = w[i];
				d = 1;
			}
			for(j = i+1; ; j++){
				power[n+d]++;
				power[d]--;
				if(j >= rank)
					break;
				if(j < 2){
					n += 2*w[j];
					d += 2;
				}else{
					n += w[j];
					d++;
				}
			}
		}
		for(pt = D_F; *(s = *pt++); pt){	/* other root */
			for(j = n = d = 0; j < 4; j++){ 
				n0 = *s++ - '0';
				if(j < 2){
					n += 2*n0*w[j];
					d += 2*n0;
				}else{
					n += n0*w[j];
					d += n0;
				}
			}
			s = *pt++;
			for( ; ; ){
				power[n+d]++;
				power[d]--;
				if(!*s)
					break;
				j = *s++ - '0';
				if(j < 2){
					n += 2*w[j];
					d += 2;
				}else{
					n += w[j];
					d++;
				}
			}
		}
		break;

	  case TYPE_G:				/* six positive roots */
		power[w[0]+1]++;
		power[w[1]+1]++;
		power[w[0] + 3*w[1] + 4]++;
		power[4]--;
		power[2*w[0] + 3*w[1] + 5]++;
		power[5]--;
		power[w[0] + w[1] + 2]++;
		power[2]--;
		power[w[0] + 2*w[1] + 3]++;
		power[3]--;
	}
	reduce(power, max_power);
	return prod_power(power, max_power);
}

/*
 *  Weyl's dimension formula for all the classical simple Lie groups
 *	In:  the rho-shifted weight by the coefficeints for usual orthogonal bases 
 */
long dim_rep(WEYL w)
{
	int i, j, c, n;

	if((max_power = w[0] + w[1]) >= MAX_POWER)
		return ERROR;
	if(max_power < 2*rank)
		max_power = 2*rank;
	clear_int(power, max_power);

	if(f_debug & FD_SHIFTWEIGHT){
		printf("Shifted weight: ");
		for(i = 0; i < rank; i++)
			printf("%d ", w[i]);
		printf("\n");
	}
	for(j = siz0; j > 0; j--){
		for(i = j-1; i >= 0; i--){
			power[w[i]-w[j]]++;
			if(wtype != TYPE_A)
				power[w[i]+w[j]]++;
		}
	}
	if(type_bc){
		for(i = siz0; i >= 0; i--)
		power[w[i]]++;
	}
	if(wtype == TYPE_A)
		n = siz0;
	else{
		if(wtype == TYPE_B){
			fact(power, rank);				/*  *rank! 				*/
			power[2] -= rank*(rank-2);		/* / 2^{rank*(rank-2)}	*/
		}else if(wtype == TYPE_C){
			fact(power, rank);				/* x rank! 				*/
			power[2] += rank;				/* x 2^rank				*/
		}else
			power[2] -= siz0*siz0;			/* / 2^{(rank-1)*(rank-1)} */
		n = 2*rank;
		if(wtype == TYPE_D)
		n -= 2;
	}
	for(c = 1; n > 1; c++){
		power[n--] -= c;
		if(wtype != TYPE_A)
			power[n--] -= c;
	}
	reduce(power, max_power);
	return prod_power(power, max_power);
}

/* -------------------------------------------------------------------- */
/*
 *  Get simple roots whose reflections by right transform WEYL <w> longer/shorter
 *
 *  In: mode  == 0 : longer
 *            != 0 : shorter
 *  ret: 2 X number of elements byte
 *       result is written in UNIT *p (NULL terminated)    
 */
int adj_srf(WEYL w, UNIT *p, int mode)
{
	int i;
	UNIT *pt;

	pt = p;
	for(i = 1; i < rank; i++){
		if(w[i-1] < w[i]){
			if(!mode){
				*pt++ = i;
				*pt++ = i+1;
			}
		}else{
			if(mode){
				*pt++ = i;
				*pt++ = i+1;
			}
		}
	}
	if(wtype == TYPE_B || wtype == TYPE_C){
		if(is_neg(w[siz0])){
			if(mode){
				*pt++ = rank;
				*pt++ = MOD0;
			}
		}else{
			if(!mode){
				*pt++ = rank;
				*pt++ = MOD0;
			}
		}
	}else if(wtype == TYPE_D){
		if(w[siz0-1] + w[siz0] < MOD){
			if(!mode){
				*pt++ = siz0;
				*pt++ = to_neg(rank);
			}
		}else{
			if(mode){
				*pt++ = siz0;
				*pt++ = to_neg(rank);
			}
		}
	}
	*pt = 0;
	return pt-p;
}

void print_ord(WEYL w)
{
	int i, j, m, n, p;
	AWEYL	v[MAX_SIZE], s[MAX_SIZE], t[MAX_SIZE];

	if(f_mode & (F_ORD0|F_ORD1)){
		if(f_mode & F_ORD1)
			inv_w(w,w);
		adj_srf(w, r_exp, f_mode & F_NEG);
	}
	else if(!(f_mode & F_NEG)){
		longest_w(s);
		mul_w(v, s, w);
		adj_br(v, r_exp);
	}else
		adj_br(w, r_exp);
	if(f_mode & F_ROOT)
		print_r(r_exp);
	else{
		for(i = 0; (m = r_exp[i++]) != 0; ){
			n = r_exp[i++];
			for(j = 0; j <= rank; j++)
				t[j] = w[j];
			if(n < MOD0){
				p = t[m-1];
				t[m-1] = t[n-1];
				t[n-1] = p;
			}else if(n == MOD0){
				t[m-1] = to_neg(t[m-1]);
			}else{
				p = t[m-1];
				n = to_abs(n);
				t[m-1] = to_neg(t[n-1]);
				t[n-1] = to_neg(p);
			}
			if(f_mode & F_ORD1)
				inv_w(t,t);
			print_e(t);
			printf(" ");
		}
	}
	if(f_mode & F_ORD1)
		inv_w(w,w);
}

void GetDebugOption(char *para)
{
	int i, ch;

	if(*para == '?'){
		printf(msg_debug);
		exit(1);
	}
	if(!*para){
		f_debug = -1;
		return;
	}
	if(*para=='X'){
		for(i = 1; i < 4 && para[i]; i++){
			if((ch = para[i]) >= '0' && ch <= '9')
				ch -= '0';
			else
				ch = (ch | 0x20) - ('a'-10);
			f_debug = f_debug*0x10 + ch; 
		}
	}else{
		for(i = 0; i < 31 && para[i]; i++){
			if(para[i] != '0')
				f_debug |= (1<<i);
		}
	}
}

int main(int argc, char *argv[])
{
	int i, j, k, pt, x, y, z;

	AWEYL w[MAX_SIZE], w2[MAX_SIZE], t[MAX_SIZE];
	WEYL	p;

	for(i = 1; i < argc; i++){						/* analyze option -???? */
		if(argv[i][0] != '-')
			break;
		if(!strncmp(argv[i]+1, "debug", 5)){
			GetDebugOption(argv[i]+6);
			continue;
		}
		for(j = 1; argv[i][j]; j++){
			switch(argv[i][j]){
				case 'W':	f_mode &= ~F_TABL;
							break;
				case 'Y':	f_mode |= F_TABL;
							break;
				case 'L':	f_mode |= F_LINEAR;
							break;
				case 'A':	wtype = TYPE_A;
							break;
				case 'B':	wtype = TYPE_B;
							break;
				case 'C':	wtype = TYPE_C;
							break;
				case 'D':	wtype = TYPE_D;
							break;
				case 'E':	wtype = TYPE_E;
							break;
				case 'F':	wtype = TYPE_F;
							break;
				case 'G':	wtype = TYPE_G;
							break;
				case 'd':	f_mode |= F_RANK;
							k = argv[i][j+1];
							if(k >= '1' && k <= '9'){
								rank = atoi(argv[i]+(++j));
								while(argv[i][j] >= '0' && argv[i][j] <= '9')
									j++;
								j--;
							}
							break;
				case 's':	f_mode |= F_SHORT;
							break;
				case 'a':	f_mode |= F_ALL;
							break;
				case 'b':	f_mode |= F_ORD;
							if(argv[i][j+1] == '0'){
								j++;
								f_mode |= F_ORD0;
							}else if(argv[i][j+1] == '1'){
								j++;
								f_mode |= F_ORD1;
							}
							break;
				case 'c':	f_mode |= F_CELL;
							if(argv[i][j+1] == '2'){
								j++;
								f_mode |= F_2CELL;
							}
							break;
				case 'w':	f_mode |= F_COSET;
							break;
				case 'o':	f_mode |= F_OPTION;
							break;
				case 'p':	f_mode |= F_PDECOMP;
							break;
				case 'r':	f_mode |= F_ROOT;
							break;
				case 'l':	f_mode |= F_LONG;
							break;
				case '-':	f_mode |= F_NEG;
							break;
				case '*':	f_mode |= F_MUL;
							break;
				case 'S':   if((s_sep[0] = argv[i][++j]) == 0)
								--j;
							break;
				case '?':	switch(argv[i][++j]){
								case 'W':	printf(msgw);
											return 0;
								case 'Y':	printf(msgy);
											return 0;
								case 'L':	printf(msgl);
											return 0;
								case 'D':
									switch(argv[i][++j]){
										case 'A':
											printf(msgla);
											return 0;
										case 'B':
											printf(msglb);
											return 0;
										case 'C':
											printf(msglc);
											return 0;
										case 'D':
											printf(msgld);
											return 0;
										case 'E':
											printf((argv[i][j+1]>'6')?
												msgle7:msgle6);
											return 0;
										case 'F':
										case 'G':
											printf(msglf);
											return 0;
										case 'R':
											printf(msgdr);
											return 0;
										default:
											printf(msgd);
											return 0;
									}
							}
			}
		}
	} 
	if(i >= argc && !rank){				/* no <exp> and rank are not given */
		if(wtype == TYPE_F)
			rank = 4;
		else if(wtype == TYPE_G)
			rank = 5;
		else{
			msg();
			return 0;
		}
	}
	if(rank || (f_mode & F_RANK)){		/* indicate by the rank */
		if(!rank)
			rank = atoi(argv[i]);
		if(rank <= 1 || rank > MAX_rRANK)
			err_exit(ER_BIGRANK);
		set_rank(rank);
		if(f_mode & F_TABL){					/* Young Tableau */
			if(s_sep[0] == 0)
				s_sep[0] = ' ';
			if(!(f_mode & (F_OPTION|F_CELL|F_SHORT)))
				print_yd(rank);
			else{
				t[0] = rank;
				t[1] = 0;
				for(i = 0; i < rank; i++)
					w[i] = i+1;
				do{
					x = dim_yd(t);
					if(x <= 0 || x >= 1000)
						err_exit(ER_BIGRANK);
					if(f_mode & F_OPTION){
						printf("\n ");
						print_e(t);
						printf("\n");
						print_2cell(t);
					}
					else for(i = 0, y = next_yt(w, t, 0); !y && i < x; i++, y = next_yt(w, t, 1)){
						if(f_mode & F_SHORT){
							if(s_sep[0] != ' '){
								print_yt(t,w);
								printf(":\n ");
							}
							adj_cell(t, w, (f_mode & F_NEG)?1:0);
							printf("\n");
						}else{
							if(s_sep[0] == ' ')
								printf("(");
							print_yt(t,w);
							printf( (s_sep[0]==' ')? ") ":" ");
						}
					}
				}while(next_yd(t) > 0);
			}
		}else if(f_mode & F_LINEAR){				/* -L option */
			t[0] = (argc > i)?atoi(argv[i]):1;
			for(j = 1; j <= rank; j++)
				t[j] = 0;
			k = rank;
			do{
				for(j = 0; j <= rank; j++)
					w[j] = t[j];
				if(wtype == TYPE_A){	/* skip the duplication of symmetry */		
					for(j = 0; j + j < rank; j++){
						if(w[j] < w[rank-1-j])
							goto nxt;
						else if(w[j] > w[rank-1-j])
							break;
					}
				}else if(wtype == TYPE_D){
					if(w[rank-2] < w[rank-1])
							goto nxt;
				}else if(wtype == TYPE_E && rank == 6){
					if(w[0] < w[5])
						goto nxt;
					if(w[0] == w[5] &&  w[1] < w[4])
						goto nxt;
				}
				printf("dim [");
				print_ew(w, rank);
				printf("] = ");
#if	0
				coef2w(w, rank);	/* coefficients -> weight, rank changes if TYPE_A */
				add_rho(w);
				dim_rep(w);
				print_power(power, 0);
				rank = k;
#else
				rank = k;
				dim_rep2(w);
				print_power(power, 0);
#endif
nxt:		;}while(next_sum(t, rank) == 0);
		}else{						/* Weyl group */
			if(wtype >= TYPE_E)
				err_exit(ER_EXCEPT);
			set_rank(rank);
//			if(f_mode & F_ORD)
				print_wo();
//			else
//				print_wy(rank);
		}
		return 0;
	}else{							/* analyze expression <exp> */
		p = &w[0];
two:	for(pt = k = x = y = z = 0; (j = argv[i][k]) != 0; k++){
			if(j >= '0' && j <= '9'){
				p[x] = j - '0';
				if(z){
					p[x] = to_neg(p[x]);
					z = 0;
				};
				x++;
			}else if(j >= 'a' && j <= 'z'){
				p[x] = j - ('a' - 10);
				if(z){
					p[x] = to_neg(p[x]);
					z = 0;
				};
				x++;
			}else if(j == '('){					/* given by (..) */
				if(!argv[i][++k])
					break;
				p[x] = atoi(argv[i]+k);
				if(p[x] < 0)
					p[x] = to_neg(-p[x]);
				x++;
				while(argv[i][++k] >= '0' && argv[i][k] <= '9');
				if(!argv[i][k])
					break;
			}else{
				z = 0;
				switch(j){
				case '/':	t[y++] = x - pt;
							pt = x; 
							break;
				case '-':	z = 1;
							break;
				case 'V':	j = 'T';
				case 'P':
				case 'Q':
				case 'R':
				case 'S':
							if(!k){
								p[x++] = ('P'-1) - j;
								break;
							}
				default:
					fprintf(stderr, "Error in <exp>:%s\n", argv[i]);
					exit(1);
				}
			}
		}
		p[x] = 0;
		if(p[0] < 0){			/* P, Q, R, S, V is at the head */
			parab_w(p, p+1, -1-p[0]);
			x = strlen(p);
		}
		if(pt){					/* exists / */
			if(x > pt)
				t[y++] = x - pt;
			t[y] = 0;
			f_mode |= F_GETTAB;
		}
		if(p == w && ++i < argc){
			p = &w2[0];
			f_mode |= F_TWO;
			goto two;
		}
	}
	p[k=x] = 0;
	if((f_mode & (F_COSET|F_TWO)) == (F_COSET|F_TWO)){
		dcoset(w,w2, f_mode&F_NEG);
	}else if((f_mode & (F_PDECOMP|F_TWO)) == (F_PDECOMP|F_TWO)){
		ParaDecomp(w, w2, f_mode&F_NEG);
	}else if(f_mode & F_GETTAB){			/* Young tableau */
		if(check_yt(t, w) < 0 || ((f_mode & F_TWO) && check_yt(t,w2)))
			err_exit(ER_BADYT);
		for(i = rank = 0; t[i]; i++)
			rank += t[i];
		set_rank(rank);
		if(f_mode & F_TWO)
			print_rs2w(t, w2, w);
		else if((f_mode & (F_SHORT|F_CELL)) == F_SHORT)
			adj_cell(t, w, (f_mode & F_NEG)?1:0);
		else if(f_mode & F_CELL)
			print_cell(t, w);
		else
			print_tab(t);
	}else if(f_mode & F_TABL){		/* Young diagram */
		k = check_yd(w);
		if(k < 0)
			err_exit(k);
		if(f_mode & F_CELL)
			print_2cell(w);
		else
			print_y(w);
	}else if(f_mode & F_LINEAR){	/* simple Lie group */
		printf("dim [");
		print_ew(w, k);
		printf("] = ");
#if	0
		coef2w(w, k);				/* coefficients -> weight */
		add_rho(w);
		dim_rep(w);
#else
		rank = k;
		dim_rep2(w);
#endif
		print_power(power, 0);
	}else{							/* element of Weyl group is given */
		set_rank(k);
		if(f_mode & F_OPTION){		/* -o expression */
			if(rank >= 10)
				err_exit(ER_BIGRANK);
			for(i = 0; i < rank; i++){
				if(w[i] >= 10 && w[i] <= 20)
					w[i] = to_neg(w[i] - 9);
			}
			if(f_mode & F_TWO){
				for(i = 0; i < rank; i++){
					if(w2[i] >= 10 && w[i] <= 20)
						w2[i] = to_neg(9 - w2[i]);
				}
			}
		}
		if(wtype >= TYPE_E)		/* exceptional Weyl group is not supported */
			err_exit(ER_EXCEPT);
		if(check_wy(w) <= 0 || ((f_mode & F_TWO) && check_wy(w2) <= 0))
			err_exit(ER_BADWY);	/* check the correctness of the expression */
		if((f_mode & F_CELL) && wtype != TYPE_A)
			err_exit(ER_CELLA);	/* cell calculation is supported only for TYPE_A */

		if((f_mode & (F_TWO|F_MUL)) == (F_TWO|F_MUL)){	/* -* option */
			mul_w(w, w, w2);
			f_mode &= ~(F_TWO|F_MUL);
		}
		if(f_mode & F_ALL){
			printf("\tAll the reduced expressions\n");
			get_r_exp(w,1);			/* get all reduced expression */
		}
		if((f_mode & (F_ROOT|F_ORD|F_CELL)) == F_ROOT){
			printf("\tRoots changed to negative\n");
			print_root(w);
			printf("\n");
		}
		if(f_mode & F_TWO)
			print_cmp(w, w2);		/* compare Bruhat ordering */
		if((f_mode & (F_ORD|F_CELL)) == F_ORD)
			print_ord(w);			/* show adjacent elements */
		if(f_mode & F_CELL)			/* left/right/two sided cell */
			print_wcell(w);
		if(!(f_mode & (F_ALL|F_ROOT|F_TWO|F_ORD|F_CELL)))
			print_w(w);				/* Robinson-Schensted etc. */
	}
	return 0;
}

/* -----------------------------------------------------------------
 *   Check if an element of weyl group/Young diagram/Young tableau
 *   is correct.
   int  check_wy(WEYL);
   int  check_yd(YD y);
   int  check_yt(YD y, YT t);
 * ----------------------------------------------------------------- */

/*
 *   Check if the expression of the element of Weyl group is correct.
 *   Also change into the innner format if not so (Type BCD)
 *       negative  2*rank+1-(n)  ->  toneg(n)
 */
int check_wy(WEYL w)
{
	int i, j, len, neg;
	AWEYL t[MAX_SIZE];

	len = strlen(w);
	if(len != rank)
		return ER_RANK;
	for(i = 0; i < len; i++)
		t[i] = 0;
	for(i = neg = 0; i < len; i++){
		j = w[i];
		if(j > len){
			if(!wtype)
				return ER_BADWY;
			if(is_neg(j)){
				j = to_neg(j);
				if(j <= 0)
					return ER_BADWY;
			}else{
				j = rank2 - j;
				if(j <= 0)
					return ER_BADWY;
				w[i] = to_neg(j);
			}
			t[j] = 1;
			neg++;
		}
		t[j] = 1;
	}
	if(wtype == TYPE_D && neg & 1)
		return ER_BADWY;
	for(i = 1; i <= len; i++)
		if(!t[i])
			return ER_BADWY;
	return len;
}

/*
 *  Check if the expression Young Diagram <y> is correct.
 *  (the sequence should be a decreasing sequence of non-negative integers)
 *  
 */
int check_yd(YD y)
{
	int	i, len;

	len = strlen(y);
	if(len > 255)
		return ER_LONG;
	for(i = len - 2; i >= 0; i--){
		if(y[i] < y[i+1] || y[i+1] < 0)
			return ER_BADYD;
	}
	return len;
}

/*
 *  Check if the Young tableau (<y>, <tab>) is standard.
 *  1. Check if Young diagram <y> is correct.
 *  2. Check if the numbers in boxes are correctly arranged.
 *
 */
int check_yt(YD y, YT tab)
{
	int i, j, k, m, n, f, len;
	UNIT w[MAX_rank];

	k = check_yd(y);
	if(k < 0)
		return k;
	for(i = len = 0; i < k; i++)
		len += y[i];
	if(len != strlen(tab))
		return ER_BADYT;
	for(i = 0; i <= len; i++)
		w[i] = 0;
	for(i = j = 0; i < len; i += k, j++){
		m = y[j];
		for(k = 0; k < m; k++){
			n = tab[i+k];
			if(n <= 0 || n > len){
				return ER_BADYT;
			}
			if(k > 0 && n <= tab[i+k-1])	/* compare with left */
				return ER_BADYT;
			if(j > 0 && n <= tab[f+k])		/* compare with upper */
				return ER_BADYT;
			w[n]++;
		}
		f = i;
	}
	for(i = 1; i <= len; i++)
		if(w[i] != 1)
			return ER_BADYT;
	return ER_OK;
}

/* ----------------------------------------------------------------- *
 *             Calculate the powers of positive integers

    void reduce(POWER *, int);
    long prod_power(POWER *, int);
	int prtprod_power(POWER *, int, int, char *);
	long fprod_power(POWER *, int, float *);
    void clear_int(POWER *, int);
	int check_int(POWER *, int);
    void fact(POWER *, int );
    void dfact(POWER *, int);
	void print_power(POWER *, int);
 * ----------------------------------------------------------------- */
/* 
 *  Reduce the product of an integer into prime factors
 *  number = 2^c[2] * 3^c[3] * .... * mx^c[mx]
 */
void reduce(POWER *c, int mx)
{
	int n, mul, j, q;

	for(n = mx; n > 3; n--){
		if(!(mul = c[n]))
			continue;
		if(!(n & 1)){					/* even number */
			c[2] += mul;				/* devided by 2 */
			c[n/2] += mul;
			c[n] = 0;
		}else{
			for(j = 3; ;j += 2){
				q = n/j;
				if(q < j)
					break;
				if(q*j == n){
					c[j] += mul;		/* devided by j */
					c[q] += mul;
					c[n] = 0;
					break;
				}
			}
		}
	}
}

/*
 *   Calculate the product as long integer
 *   ret:   > 0  the product
 *           -1  overflow
 */
long prod_power(POWER *pow, int mx)
{
	int i, j;
	long d;

	d = 1 << pow[2];
	for(i = 3; i <= mx; i += 2){
		for(j = power[i]; j > 0; j--){
			d *= i;				/* *= i^pow[i] */
			if(d < 0)
				return -1;
		}
	}
	return d;
}

/*
 *   Print the product as an integer
 *   In: <keta>   Display the number if it is less than 10^<keta>
 *   ret:  0      Ok
 *        -1      Overflow
 */
int prtprod_power(POWER *pow, int mx, int keta, char *buf)
{
	int i, j, k, n, kp, kmx;
	int *p;

#define	KETA_UNIT	10000	/* KETA_UNIT*KETA_UNIT shoud be handled as int */
#define	KETA_FIG	"4"
#define	KETA_NUM	 4

	kmx = keta/4 + 1;
	p = (int *)malloc(sizeof(int)*kmx);
	*p = 1;
	for(i = 1; i < kmx; i++)
		p[i] = 0;

	for(kp = 1, i = 2; i <= mx; i++){
		for(j = pow[i]; j > 0; j--){
			for(n = k = 0; k < kp; k++){
				p[k] =  p[k]*i + n;
				if(p[k] >= KETA_UNIT){
					n = p[k]/KETA_UNIT;
					p[k] -= n*KETA_UNIT;
				}else
					n = 0;
			}
			if(n){
				if(kp++ >= kmx || n >= KETA_UNIT){
					free(p);
					return -1;
				}
				p[k] = n;
			}
		}
	}
	if(buf == NULL){
		printf("%d", p[kp-1]);
		for(k = kp-2; k >= 0; k--)
			printf("%0" KETA_FIG "d", p[k]);
	}else{
		sprintf(buf, "%d", p[kp-1]);
		buf += strlen(buf);
		for(k = kp-2; k >= 0; k--, buf += KETA_NUM)
			printf("%0" KETA_FIG "d", p[k]);
	}
	free(p);
	return 0;
}

/*
 *  Get the product as a decimal number
 *  Out: <f>  floating number between 1.0000 and 9.9999 by cutting lower terms
 *       ret  exponent of the powe of 10
 *
 *  the product = <f> x 10^<ret>
 */
long fprod_power(POWER *pow, int mx, float *f)
{
	int i, j, k;
	float d;

	k = 0;
	d = 1;
	for(i = 2; i <= mx; i++){
		for(j = pow[i]; j > 0; j--){
			d *= i;				/* *= i^pow[i] */
			if(d >=  1000000.0){
				d /= 1000000.0;
				k += 6;
			}
		}
	}
	while(d >= 10){
		d /= 10;
		k++;
	}
	d = floor(d * 10000)/10000;
	*f = d;
	return k;
}

/*
 *    Clear the work space of the calculation
 */
void clear_int(POWER *c, int mx)
{
	int i;

	for(i = 0; i <= mx; i++)
		c[i] = 0;
}

/*
 *   Check if there exists only non-negative powers of integers
 */
int check_int(POWER *pow, int mx)
{
	int	i;

	for(i = mx; i >= 2; i--)
		if(pow[i] < 0)
			return ERROR;
	return 0;
}

/*
 *     multiply by  <n>!
 */
void fact(POWER *pow, int n)
{
	int i;

	for(i = n; i > 1; i--)
		pow[i]++;
}

/*
 *     devide by  <n>!
 */
void dfact(POWER *pow, int n)
{
	int i;

	for(i = n; i > 1; i--)
		pow[i]--;
}

/*
 *  A sample to print the product
 */
void print_power(POWER *pow, int mx)
{
	int i, p, flag;
	float f;

#define	MAX_KETA	100

	if(!mx)
		mx = max_power;
	p = prod_power(pow, mx);
	if(p > 0)
		printf("%ld", p);
	else{
		i = fprod_power(pow, mx, &f);
		if(i <= MAX_KETA){
			prtprod_power(pow, mx, i, NULL);
			printf(" = ");
		}
		printf("%.4f...*10^%ld", f, i);
	}
	for(flag = 0, i = 2; i < mx; i++){
		if( (p = pow[i]) > 0){
			if(!flag){
				printf(" = ");
				flag = 1;
			}
			printf("(%d", i);
			if(p>1)
				printf("^%d)",p);
			else
				printf(")");
		}
	}
	printf("\n");
}


/* ---------------------------------------------------------------------- */

/* 
 *  Calculate the dimension of the representation expressed by Yong diagram
 *
 *  In   w: Young Diagram
 *  Ret     dimension
 *          negative value -> error
 */
long dim_yd(YD w)
{
	int	i, j, len, d, raw;

	len = check_yd(w);
	power[2] = 0;
	if(len < 0)
		return len;
	for(i = d = 0; i < len; i++)
		d += w[i];					/* degree of YT */
	if(d > 255)
		return ER_LONG;
	for(i = 1; i <= d; )
		power[i++] = 1;
	for(j = 0; j < len; j++){
		for(i = 1, raw = len-1; i <= w[j] ; i++){
			while(w[raw] < i)		/* bottom raw of (i,j+1) */
				raw--;
		if(f_debug & FD_HOOKLENGTH)
			printf("(%d,%d)=%d\n", j+1, i, w[j]  - i + raw - j + 1);
												/* hook length */
			power[w[j] - i + raw - j + 1]--;
		}
	}
	reduce(power, d);
	if(check_int(power, d))
		return ER_SYSTEM;
	return prod_power(power, d);
}


/*
 *   Calculate the inverse of Robinson-Schensted correspondence
 *
 *   In   y: Young diagram
 *       (ly, ry): a pair of Youg tableau
 *   ret  w: an element of symmetric group(permutation)
 */
int rs2w(WEYL w, YD y, YT ly, YT ry)
{
	int i, j, num, pt, raw, max_raw;
	UNIT yt[MAX_SIZE];
	UNIT pos[MAX_SIZE];
	UNIT sumy[MAX_SIZE];

	if(f_debug & FD_IRS){
		printf("\nInverse of Robinson-Schensted:");
		print_yt(y, ly);
		printf(" ");
		print_yt(y, ry);
		if( (i = check_yt(y, ly)) < 0 || (i = check_yt(y, ry)) < 0){
			printf("\n");
			err_exit(i);
		}
	}
	w += rank;
	*w-- = 0;
	for(i = 0; i < rank; i++){
		pos[ly[i]] = i;
		yt[i] = ry[i];
	}
	max_raw = get_sumyd(sumy, y);
	yt[i] = 0;
	while(i > 0){
		num = i--;
		pt =  pos[num];
		num = yt[pt];
		yt[pt] = 127;									/* erase box */
		for( raw = max_raw; pt < sumy[--raw] ; );		/* search raw */ 
		while(raw > 0){
			for(pt = sumy[raw]; yt[--pt] > num; );		/* last box smaller than num */
			j = num;
			num = yt[pt];
			yt[pt] = j;
			raw--;
		}
		*w-- = num;
	}
	if(f_debug & FD_IRS){
		printf(" -> ");
		print_e(w+1);
		printf("\n");
	}
	return 0;
}


/*
 *   Calculate the Robinson-Schensted correspondence
 *
 *   In     w: an element of symmetric group(permutation) 
 *   Ret:   y: Young Diagram
 *         ry, ly: pair of Young tableau
 */
int rs(WEYL w, YD y, YT ly, YT ry)
{
	int i, l, n, m, p, c, t;

	if(f_debug & FD_RS){
		printf("\nRobinson-Schensted: ");
		print_e(w);
		printf(" -->\n\n");
	}
	*y = *ry = *ly = 0;
	l = strlen(w);
	if(wtype == TYPE_C || wtype == TYPE_D)
		l += l;
	else if(wtype == TYPE_B)
		l += l+1;
	if(l >= MAX_SIZE)
		return ER_LONG;
	for(i = l; i >= 0; i--)
		y[i] = ry[i] = ly[i] = 0;
	i = 0;
	if(wtype == TYPE_B)
		y[0] = 1;

	while((n = w[i++]) != 0){
		if(wtype == TYPE_A){
			for(p = l = 0; n != 0; l++){
				c = y[l];
				while(c--){
					if(n > ry[p++])
						continue;
					t = ry[p-1];				/* swap number */
					ry[p-1] = n;
					n = t;
					p += c;
					break;
				}
				if(c < 0){						/* add a box to a row */
					y[l]++;
					for(t = i; t > p; t--){
						ry[t] = ry[t-1];		/* move for the insertion */
						ly[t] = ly[t-1];
					}
					ry[p] = n;					/* insert */
					ly[p] = i;
					n = 0;
				}
			}
		}
#if 0
         else{
			int j;

			m = to_abs(n);
			f_tate = (m == n)?0:1;
			l = 0;

			if(!f_tate){
yok0:			for(j = p = 0; j < l; i++)
					p += y[j];
				c = y[l];
				while(c--){
					if(m > ry[p++])
						continue;
					t = ry[p-1];				/* swap number */
					ry[p-1] = m;
					m = t;
					break;
				}
				if(c < 0){						/* add two box to a row */
					y[l] += 2;
					for(t = 2*i+1; t > p; t--){
						ry[t] = ry[t-2];		/* move for the insertion */
						ly[t] = ly[t-2];
					}
					ry[p] = ry[p+1] = m;		/* insert */
					ly[p] = ly[p+1] = i;
					continue;
				}else if(c == 0){				/* add a box to a row */
					y[l+1]++;					/* y[l] == y[l+1] */
					y[l]++;
					n = p + y[l];
					for(t = 2*i + 1; t > n; t--){
						ry[t] = ry[t-2];		/* move for the insertion */
						ly[t] = ly[t-2];
					}
					ry[p] = m;
					ry[n] = ry[n-2];
					ly[p] = ly[n] = i;
					for(t = n-1; t > p; t--){
						ry[t] = ry[t-1];		/* move for the insertion */
						ly[t] = ly[t-1];
					}
					continue;
				}else{
					n = ry[p];
					ry[p] = ry[p-1];
					if(m == n){
						l++;
						goto yok0;
					}
yok1:				j = y[l] - c;				/* j-th column <- p */
					if(j >= y[l+1]){
						p += c + y[l+1];				/* p + c: next top */
						for(t = 2*i-1; t >= p; t--){
							ry[t+2] = ry[t];	/* move for the insertion */
							ly[t+2] = ly[t];
						}
						y[l+1] += 2;
						ry[p] = m;
						ry[p+1] = n;
						ly[p] = ly[p+1] = i;
						continue;
					}
					t = ry[p+c+j];
					ry[p+c+j] = m;
					m = n;
					n = t;
					t = j;
					j = l;
					l = t;
					goto tat1;
				}
			}else{
tat0:			for(p = l, j = 0; ; p += y[j++]){
					if(y[j] <= l || m < ry[p])
						break;
				}
				if(y[j] <= l){
					for(p = t = 0; t <= j+1; t++)
						p += y[t];
					for(t = 2*i + 1; t >= p; t--){
						ry[t+2] = ry[t];
						ly[t+2] = ly[t];
					}
					ry[p+1] = m;
					ly[p+1] = i;
					p -= y[j];
					y[j]++;
					y[j+1]++;
					for( ; t >= p; t--){
						ry[t+1] = ry[t];
						ly[t+1] = ly[t];
					}
					ry[p] = m;
					ly[p] = i;
					continue;
				}
				if(y[j+1] <= l){
					n = ry[p];
					ry[p] = m;
					for(p = t = 0; t <= j+1; t++)
						p += y[t];
					for(t = 2*i + 1; t >= p; t--){
						ry[t+2] = ry[t];
						ly[t+2] = ly[t];
					}
					ry[p] = m;
					ry[p+1] = n;
					ly[p] = ly]p+1] = i;
					y[j+1] += 2;
					continue;
				}else{
					t = ry[p]:
					n = ry[p+y[j]];
					ry[p+y[i]] = ry[p] = m;
					m = t;
					if(m == n){
						l++;
						goto tat0;
					}
				}
tat1:			if(y[j] <= l+1){
					for(p = t = 0; t <= j+2; t++)
						p += y[t];
					for(t = 2*i + 1; t >= p; t--){
						ry[t+2] = ry[t];
						ly[t+2] = ly[t];
					}
					ry[p+1] = n;
					ly[p+1] = i;
					p -= y[j+2];
					for(; t >= p; t--){
						ry[t+1] = ry[t];
						ly[t+1] = ly[t];
					}
					ry[p] = m;
					ly[p] = i;
					y[j+1]++;
					y[j+2]++;
					continue;
				}else{
					t = ry[p+y[j]+1]
					ry[p+y[j]+1] = m;
					m = n;
					n = t;
					c = y[j+1] - l - 1;
					l = j+1;
					goto yok1;
				}
			}
		}
#endif
		if(f_debug & FD_RS){
			int xx, yy, zz;

			for(xx = zz = 0; (yy = y[xx++])!= 0; ){
				while(yy--)
					printf("%3d", ry[zz++]);
				printf("\n");
			}
			printf("\n");
			for(xx = zz = 0; (yy = y[xx++])!= 0; ){
				while(yy--)
					printf("%3d", ly[zz++]);
				printf("\n");
			}
			printf("\n\n");
		}
	}
	return ER_OK;
}


/*
 *  Get the representative with the minimal length in the given right coset 
 *  by  a parabolic subgroup.
 *
 *  ret  1 : change into the element with the minimal length
 *       0 : the given element is of the minimal length
 */
int repright(WEYL w, YD y)
{
	int i, j, k, m, n, flag, tmp, fd;
	UNIT wt[MAX_rank];

	flag = m = n = fd = 0;
	for( ; ; m = n){			/* sort w[m],...,w[n-1] */
		n += *y++;
		if(n > rank){			/* type_B/C/D */
			if(wtype == TYPE_D){
				j = -1;
				for(i = m+1; i < n; i++){
					if(is_neg(w[i])){
						if(j >= 0){
							w[i] = to_neg(w[i]);
							w[j] = to_neg(w[j]);
							flag = 1;
							j = -1;
						}else
							j = i;
					}
				}
				if(j >= 0)
					fd = 1;
			}
			else for(i = m+1; i < n; i++){
				if(is_neg(w[i])){
					w[i] = to_neg(w[i]);
					flag = 1;
				}
			}
			n = rank;
		}
		for(i = m+1; i < n; i++){
			for(j = i; j > m && w[j-1] > w[j]; j--){
				flag = 1;
				tmp = w[j-1];
				w[j-1] = w[j];
				w[j] = tmp;
			}
		}
		if(n >= rank){
			if(fd){
				for(fd = -1, i = m+1; i < n; i++){
					if(w[i] + w[siz0] > MOD)
						fd = i;
					else{
						if(fd >= 0){
							tmp = to_neg(w[siz0]);
							w[siz0] = to_neg(w[fd]);
							w[fd] = tmp;
							flag = 1;
						}else
							break;
					}
				}
			}
			break;
		}
	}
	return flag;
}

/*
 * Change w[] into the adjacent larger element under permutations
 * w[0],...,w[rank-1]: given numbers (some numbers may be same)
 *
 * ret 0: changed
 *     -1: given one is the largest
 */
int NextOrd(WEYL w)
{
	int i, j, tmp;

	for(i = rank - 2; i >= 0; i--){
		if(w[i] < w[i+1]){
			for(j = rank - 1; w[i] >= w[j]; j--);
			tmp = w[j];
			w[j] = w[i];
			w[i] = tmp;
			for(i++, j = rank-1; i < j; i++, j--){
				tmp = w[j];
				w[j] = w[i];
				w[i] = tmp;
			}
			return 0;
		}
	}
	return -1;
}

// #define	TEST	0

/*
 *  Generate representatives (with minimal length) of the right cosets by a 
 *  parabolic subgroup
 *  YD <y>   partition (parabolic)
 *           NULL: next representative
 *  return   1: get the next representative
 *           0: end
 */
int NextRep(WEYL w, YD y)
{
	int i, j, k;
	static int sign, mx_sign;
	static UNIT wo[MAX_rank];
	static UNIT wd[MAX_rank];

	if(y != NULL){
		if(rank > 31)
			return 0;
		wo[rank] = 0;
		for(i = 0; y[i] ; i++)
			wd[i] = y[i];
		wd[i] = 0;
		for(i = 0, k = 1; i < rank; i){
			for(j = *y++; j-- > 0; )
				wo[i++]   = k;
			k = i + 1;
		}
		k = wo[rank];
		mx_sign = 1;
		if(wtype != TYPE_A){
			for(i = rank-1; i >= 0 && wo[i] == k; i--);
			if(wtype == TYPE_D && i == rank-1)
				i--;
			for(mx_sign = 1; i-- >= 0; )
				mx_sign *= 2;
		}
		sign = 0;
	}
#ifdef	TEST
printf("\n %d %d ", sign, mx_sign);
print_el(wo, rank);
#endif
	if(sign >= mx_sign){
		if(NextOrd(wo))
			return 0;
		sign = 0;
	}
#ifdef	TEST
printf("->");
print_el(wo, rank);
#endif
	for(i = rank-1; i >= 0 ; i--){
		k = wo[i];
		for(j = i-1; j >= 0; j--)
			if(wo[j] == wo[i])
				k++;
		w[k-1] = i+1;
	}
	if(wtype != TYPE_A){
		k = 1;
		for(i = j = 0; i < rank; i++, k += k){
			if(k & sign){
				j++;
				w[i] = to_neg(w[i]);
			}
		}
		if(wtype == TYPE_D && (j & 1))
			w[rank-1] = to_neg(w[rank-1]);
		repright(w, wd);
	}
	sign++;
	return 1;
}


/*
 *  Parabolic decomposition
 */
void ParaDecomp(WEYL w, YT y, int mode)
{
	int i, j, m;
	UNIT wm[MAX_rank], wp[MAX_rank];

	for(m = i = 0; y[i]; i++)
		m += y[i];
	if(wtype == TYPE_A)
		set_rank(m);
	else
		set_rank(m-1);
	if( (m = check_wy(w)) < 0 )
		err_exit(m);
	if(mode)
		inv_w(wm, w);
	else
		strcpy(wm, w);
	repright(wm, y);
	if(mode){
		mul_w(wp, w, wm);
		inv_w(wm, wm);
		print_e(wp);
		printf("*");
		print_e(wm);
	}else{
		print_e(wm);
		inv_w(wm, wm);
		mul_w(wp, wm, w);
		printf("*");
		print_e(wp);
	}
	printf("\n");
}

/*
 *  Get representatives of the double cosets by parabolic subgroups
 *  mode 0: the representatives with minimal length
 *       1: the representatives with maximal length 
 */
void dcoset(YT yl, YT yr, int mode)
{
	int	i, j, m;
	UNIT wt[MAX_rank], w1[MAX_rank], w2[MAX_rank];
	UNIT ylt[MAX_rank];

	for(m = i = 0; yl[i]; i++)
		m += ylt[i] = yl[i];
	ylt[i] = 0;
	if(wtype == TYPE_A)
		set_rank(m);
	else
		set_rank(m-1);
	for(m = i = 0; yr[i]; )
		m += yr[i++];
	if(wtype == TYPE_A){
		if(m != rank)
er:			err_exit(ER_BADRANK);
	}else if(m != rank + 1)
		goto er;
	if(mode){
		longest_w(w1);
		for(i = 0; ylt[i]; i++);
		i--;
		if(wtype == TYPE_A){
			for(j = 0; j <= i; j++)
				ylt[j] = yl[i-j];
		}else if(wtype == TYPE_D && (rank & 1)){
			if(i >= 1){
				if(ylt[i] == 1 && ylt[i-1] > 1){
					ylt[i] = 2;
					ylt[i-1]--;
				}else if(ylt[i] == 2 ){
					ylt[i] = 1;
					ylt[i-1]++;
				}
			}
		}
	}
	if(NextRep(wt, yr) == 0)
		return;
	do{
#ifdef	TEST
printf("\n");
print_el(wt, rank);
#endif
		inv_w(w2, wt);
#ifdef	TEST
printf(" ");
print_el(w2, rank);
#endif
// printf("[%d %d]", ylt[0], ylt[1]);
		if(repright(w2, ylt))
			continue;
		if(mode){
//			printf("codim %2d:", length(wt));
			mul_w(w2, w1, wt);
//			print_el(w2, rank);
			add_w_buf_count(w2);
		}else{
//			printf("dim %2d:", length(wt));
//			print_el(wt, rank);
			add_w_buf_count(wt);
		}
//		printf("\n");
	}
	while(NextRep(wt, NULL));
	print_w_buf(3);
	if(!bruhat_poset() && (f_mode & F_ORD)){
		printf("\n\tadjacent Bruhat relation\n");
		for(i = 0; i < w_count; i++){
			m = power[i+1];
			j = power[i];
			if(j < m){
				print_el(w_buf + rank*i, rank);
				printf(">");
				while(j < m){
					print_el(w_buf + rank*power[--m], rank);
					printf(" ");
				}
				printf("\n");
			}
		}
	}
}
